"""Base tools to build simple buckets out of Quantitative and Qualitative features
for a binary classification model.
"""

import json
from abc import ABC
from dataclasses import dataclass

from numpy import nan, select
from pandas import DataFrame, Series
from sklearn.base import BaseEstimator, TransformerMixin

from ...combinations import CombinationEvaluator
from ...features import BaseFeature, Features
from ...utils import extend_docstring, get_attribute, get_bool_attribute
from .multiprocessing import apply_async_function


@dataclass
class Sample:
    """sample class to store X and y"""

    X: DataFrame
    y: Series = None

    def __getitem__(self, key):
        if key == "X":
            return self.X
        if key == "y":
            return self.y

        raise KeyError(key)

    def __iter__(self):
        return iter(["X", "y"])

    def keys(self):
        return ["X", "y"]

    @property
    def shape(self):
        return self.X.shape

    @property
    def index(self):
        return self.X.index

    @property
    def columns(self):
        return self.X.columns

    def __len__(self):
        return len(self.X)

    def fillna(self, features: Features) -> None:
        """fills up nans for features that have some"""
        self.X = features.fillna(self.X)

    def unfillna(self, features: Features) -> DataFrame:
        """reinstating nans when not supposed to group them"""
        return features.unfillna(self.X)


class BaseDiscretizer(ABC, BaseEstimator, TransformerMixin):
    """Applies discretization using a dict of GroupedList to transform a DataFrame's columns.

    Examples
    --------
    See `Discretizers examples <https://autocarver.readthedocs.io/en/latest/index.html>`_
    """

    __name__ = "BaseDiscretizer"

    def __init__(
        self,
        features: Features,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------

        features : Features
            A set of :class:`Features` to be processed

        min_freq : float
            Minimum frequency per modality per feature

            * Features need at least one modality more frequent than :attr:`min_freq`
            * Defines number of quantiles of continuous features
            * Minimum frequency of modality of quantitative features

            .. tip::
                Set between ``0.01`` (slower, less robust) and ``0.2`` (faster, more robust)

        Keyword Arguments
        -----------------

        ordinal_encoding : bool, optional
            Whether or not to ordinal encode :class:`Features`, by default ``False``

        copy : bool, optional
            Copying input data, by default ``False``

        verbose : bool, optional
            * ``True``, without ``IPython``: prints raw statitics
            * ``True``, with ``IPython``: prints HTML statistics, by default ``False``

        n_jobs : int, optional
            Processes for multiprocessing, by default ``1``
        """
        # features and values
        self.features = features
        if isinstance(features, list):
            self.features = Features(features)

        # saving kwargs
        self.kwargs = kwargs

        # checking types of bool attributes
        self.copy = get_bool_attribute(kwargs, "copy", True)
        self.ordinal_encoding = get_bool_attribute(kwargs, "ordinal_encoding", False)
        self.dropna = get_bool_attribute(kwargs, "dropna", False)
        self._verbose = get_bool_attribute(kwargs, "verbose", True)

        # setting number of jobs
        self.n_jobs = get_attribute(kwargs, "n_jobs", 1)

        # check if the discretizer has already been fitted
        self.is_fitted = get_bool_attribute(kwargs, "is_fitted", False)

        # carver attributes
        self.min_freq = kwargs.get("min_freq")  # default to None
        self.combinations = kwargs.get("combinations")  # default to None

    def __repr__(self, N_CHAR_MAX: int = 700) -> str:
        """Returns the string representation of the Discretizer"""
        _ = N_CHAR_MAX  # unused attribute
        # truncating features if too long
        str_features = str(self.features)
        if len(str_features) > N_CHAR_MAX:
            str_features = str_features[:N_CHAR_MAX] + "..."
        return f"{self.__name__}({str_features})"

    @property
    def categoricals(self) -> list[str]:
        """Returns the list of categorical features"""
        return self.features.categoricals

    @property
    def quantitatives(self) -> list[str]:
        """Returns the list of quantitative features"""
        return self.features.quantitatives

    @property
    def qualitatives(self) -> list[str]:
        """Returns the list of qualitative features"""
        return self.features.qualitatives

    @property
    def ordinals(self) -> list[str]:
        """Returns the list of ordinal features"""
        return self.features.ordinals

    @property
    def verbose(self) -> bool:
        """Returns the verbose attribute"""
        return self._verbose

    @verbose.setter
    def verbose(self, value: bool) -> None:
        """Sets the verbose attribute"""
        self._verbose = value

    def _remove_feature(self, feature: str) -> None:
        """Removes a feature from all ``BaseDiscretizer.feature`` attributes

        Parameters
        ----------
        feature : str
            Column name of the feature to remove
        """
        if feature in self.features:
            self.features.remove(feature)

    def _cast_features(self, X: DataFrame) -> DataFrame:
        """Casts the features of a DataFrame using feature versions to duplicate columns

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in
            ``BaseDiscretizer.features``, by default None.

        Returns
        -------
        DataFrame
            A formatted X
        """

        # duplicating columns that have several versions
        casted_columns = {
            feature.version: X[feature.name]
            for feature in self.features
            if feature.version != feature.name and feature.version not in X
        }

        # duplicating features with versions disctinct from names (= multiclass target)
        if len(casted_columns) > 0:  # checking for casted feature to not break inplace
            X = X.assign(**casted_columns)

        return X

    def _prepare_y(self, y: Series) -> None:
        """Validates input y"""

        if not isinstance(y, Series):  # checking for y's type
            raise ValueError(f"[{self.__name__}] y must be a pandas.Series, passed {type(y)}")

        if any(y.isna()):  # checking for nans in the target
            raise ValueError(f"[{self.__name__}] y should not contain numpy.nan")

    def _prepare_X(self, X: DataFrame) -> DataFrame:
        """Validates input X"""

        # checking for X's type
        if not isinstance(X, DataFrame):
            raise ValueError(f"[{self.__name__}] X must be a pandas.DataFrame, passed {type(X)}")

        # copying X
        x_copy = X
        if self.copy:
            x_copy = X.copy()

        # checking for input columns by feature name
        missing_columns = [feature for feature in self.features if feature.name not in x_copy]
        if len(missing_columns) > 0:
            raise ValueError(
                f"[{self.__name__}] Requested discretization of {str(missing_columns)} but "
                "those columns are missing from provided X. Please check your inputs! "
            )

        # casting features for multiclass targets
        x_copy = self._cast_features(x_copy)

        # checking for input columns by feature version
        missing_columns = [feature for feature in self.features if feature.version not in x_copy]
        if len(missing_columns) > 0:
            raise ValueError(
                f"[{self.__name__}] Requested discretization of {str(missing_columns)} but "
                "those columns are missing from provided X. Please check your inputs! "
            )

        return x_copy

    def _prepare_data(self, sample: Sample) -> Sample:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in
            ``BaseDiscretizer.features``, by default None.

        y : Series
            Binary target feature, by default None.

        Returns
        -------
        DataFrame
            A formatted copy of X
        """

        # checking DataFrame of features
        if sample.X is not None:
            sample.X = self._prepare_X(sample.X)

            # checking target Series
            if sample.y is not None:
                self._prepare_y(sample.y)

                # checking for matching indices
                if not len(sample.y.index) == len(sample.X.index) or not all(
                    sample.y.index == sample.X.index
                ):
                    raise ValueError(f"[{self.__name__}] X and y must have the same indices.")

        return sample

    __prepare_data = _prepare_data  # private copy

    def fit(self, X: DataFrame = None, y: Series = None) -> None:
        """Learns simple discretization of values of X according to values of y.

        Parameters
        ----------
        X : DataFrame
            Training dataset, to determine features' optimal carving
            Needs to have columns has specified in ``features`` attribute.

        y : Series
            Target with wich the association is maximized.
        """
        _, _ = X, y  # unused arguments

        # checking for previous fits of the discretizer that could cause unwanted errors
        if self.is_fitted:
            raise RuntimeError(
                f"[{self.__name__}] Already fitted. "
                "Fitting it anew could break it. Please initialize a new one."
            )

        # checking that all features were fitted
        missing_features = [feature.version for feature in self.features if not feature.is_fitted]
        if len(missing_features) != 0:
            raise RuntimeError(f"[{self.__name__}] Features not fitted: {str(missing_features)}.")

        # setting features in ordinal encoding mode
        self.features.ordinal_encoding = self.ordinal_encoding

        # setting fitted as True to raise alerts
        self.is_fitted = True

        return self

    def transform(self, X: DataFrame, y: Series = None) -> DataFrame:
        """Applies discretization to a DataFrame's columns.

        Parameters
        ----------
        X : DataFrame
            Dataset to be carved.
            Needs to have columns from provided :class:`Features`.
        y : Series, optional
            Target, by default ``None``

        Returns
        -------
        DataFrame
            Discretized X.
        """
        # * For each feature's value, the associated group label is attributed (as definid by
        # ``values_orders``).
        # * If ``ordinal_encoding="float"``, converts labels into floats.
        # * Data types are matched as ``input_dtypes=="str"`` for qualitative features and
        # ``input_dtypes=="float"`` for quantitative ones.
        # * If ``copy=True``, the input DataFrame will be copied.

        # checking that it was fitted
        if not self.is_fitted:
            raise RuntimeError(f"[{self.__name__}] Call fit method first.")

        # copying dataframes and casting for multiclass
        sample = self.__prepare_data(Sample(X, y))

        # filling up nans for features that have some
        sample.fillna(self.features)

        # checking that all unique values in X are in features
        self.features.check_values(sample.X)

        # transforming quantitative features
        if len(self.features.quantitatives) > 0:
            sample = self._transform_quantitative(sample)

        # transforming qualitative features
        if len(self.features.qualitatives) > 0:
            sample = self._transform_qualitative(sample)

        # reinstating nans when not supposed to group them
        return sample.unfillna(self.features)

    def _transform_quantitative(self, sample: Sample) -> Sample:
        """Applies discretization to a DataFrame's Quantitative columns.

        * Data types are defined by:
            * ``input_dtypes=="str"`` for qualitative features
            * ``input_dtypes=="float"`` for quantitative features
        * If ``copye=True``, the input DataFrame will be copied.

        Parameters
        ----------
        X : DataFrame
            Contains columns named after ``BaseDiscretizer.features`` attribute, by default None
        y : Series, optional
            Model target, by default None

        Returns
        -------
        DataFrame
            Discretized X.
        """
        # transforming all features
        transformed = apply_async_function(
            transform_quantitative_feature,
            self.features.quantitatives,
            self.n_jobs,
            sample.X,
            sample.shape[0],
        )

        # unpacking transformed series
        sample.X[[feature for feature, _ in transformed]] = DataFrame(
            dict(transformed), index=sample.index
        )

        return sample

    def _transform_qualitative(self, sample: Sample) -> Sample:
        """Applies discretization to a DataFrame's Qualitative columns.

        * Data types are defined by:
            * ``input_dtypes=="str"`` for qualitative features
            * ``input_dtypes=="float"`` for quantitative features
        * If ``copye=True``, the input DataFrame will be copied.

        Parameters
        ----------
        X : DataFrame
            Contains columns named after ``BaseDiscretizer.features`` attribute, by default None
        y : Series, optional
            Model target, by default None

        Returns
        -------
        DataFrame
            Discretized X.
        """
        # list of qualitative features
        qualitatives = self.features.qualitatives

        # replacing values for there corresponding label
        sample.X.replace(
            {feature.version: feature.label_per_value for feature in qualitatives}, inplace=True
        )

        return sample

    def _log_if_verbose(self, prefix: str = " -") -> None:
        """prints logs if requested"""
        if self.verbose:
            print(f"{prefix} [{self.__name__}] Fit {str(self.features)}")

    def to_json(self, light_mode: bool = False) -> str:
        """Converts to JSON format.

        To be used with ``json.dump``.

        Parameters
        ----------
        light_mode: bool, optional
            Whether or not to save features' history and statistics, by default False

        Returns
        -------
        str
            JSON serialized object
        """
        # adding all parameters
        content = {
            "features": self.features.to_json(light_mode),
            "dropna": self.dropna,
            "min_freq": self.min_freq,
            "combinations": self.combinations,
            "is_fitted": self.is_fitted,
            "n_jobs": self.n_jobs,
            "verbose": self.verbose,
            "ordinal_encoding": self.ordinal_encoding,
        }

        # adding combinations if it exists
        if isinstance(self.combinations, CombinationEvaluator):
            content["combinations"] = self.combinations.to_json()

        return content

    def save(self, file_name: str, light_mode: bool = False) -> None:
        """Saves pipeline to .json file.

        Parameters
        ----------
        file_name : str
            String of .json file name.
        light_mode: bool, optional
            Whether or not to save features' history and statistics, by default False

        Returns
        -------
        str
            JSON serialized object
        """
        # checking for input
        if file_name.endswith(".json"):
            with open(file_name, "w", encoding="utf-8") as json_file:
                json.dump(self.to_json(light_mode), json_file)
        # raising for non-json file name
        else:
            raise ValueError(f"[{self.__name__}] Provide a file_name that ends with .json.")

    @classmethod
    def load(cls, file_name: str) -> "BaseDiscretizer":
        """Allows one to load an Discretizer saved as a .json file.

        The Discretizer has to be saved with ``Discretizer.save()``, otherwise there
        can be no guarantee for it to be restored.

        Parameters
        ----------
        file_name : str
            String of saved Discretizer's .json file name.

        Returns
        -------
        BaseDiscretizer
            A fitted Discretizer.
        """
        # reading file
        with open(file_name, "r", encoding="utf-8") as json_file:
            discretizer_json = json.load(json_file)

        # deserializing features
        features = Features.load(discretizer_json.pop("features"))

        # initiating BaseDiscretizer
        return cls(features=features, **discretizer_json)

    @extend_docstring(Features.summary)
    @property
    def summary(self) -> DataFrame:
        return self.features.summary

    @property
    def history(self) -> DataFrame:
        """History of discretization process for all features"""
        return self.features.history

    # def update(
    #     self,
    #     feature: str,
    #     mode: str,
    #     discarded_value: Union[str, float],
    #     kept_value: Union[str, float],
    # ) -> None:
    #     """Allows one to update the discretization groups

    #     Use with caution: no viability checks are performed.

    #     Parameters
    #     ----------
    #     feature : str
    #         Specify for which feature to update the discritezer
    #     mode : str
    #         * For ``mode="replace"``, ``discarded_value`` will be replaced by ``kept_value``
    #         * For ``mode="group"``, ``discarded_value`` will be grouped with ``kept_value``
    #     discarded_value : Union[str, float]
    #         A group value that won't exist after completion
    #     new_value : Union[str, float]
    #         A group value that will persist after completion
    #     """
    #     # checking for mode
    #     assert mode in ["group", "replace"], " - [Discretizer] Choose mode in
    # ['group', 'replace']"

    #     # checking for nans
    #     if isnan(discarded_value):
    #         discarded_value = self.nan
    #         self.features_dropna[feature] = True
    #     assert not isnan(
    #         kept_value
    #     ), " - [Discretizer] missing values can only be grouped with an existing modality"

    #     # copying values_orders
    #     values_orders = {k: v for k, v in self.values_orders.items()}
    #     order = values_orders[feature]

    #     # checking that discarded_value is not already in new_value
    #     if order.get_group(discarded_value) == kept_value:
    #         warn(
    #             f"[Discretizer] {discarded_value} is already grouped within {kept_value}",
    #             UserWarning,
    #         )

    #     # otherwise, proceeding
    #     else:
    #         # adding kept_value if it does not exists yet
    #         if not order.contains(kept_value):
    #             order.append(kept_value)

    #         # grouping discarded value in kept_value
    #         if mode == "group":
    #             # adding discarded_value if it does not exists yet
    #             if not order.contains(discarded_value):
    #                 order.append(discarded_value)
    #             # grouping discarded_value with kept_value
    #             order.group(discarded_value, kept_value)

    #         # replacing group leader if requested
    #         elif mode == "replace":
    #             # grouping kept_value with discarded_value
    #             order.group(kept_value, discarded_value)

    #             # checking that kept_value is in discarded_value
    #             assert order.get_group(kept_value) == discarded_value, (
    #                 f"[Discretizer] Can not proceed! {kept_value} already is a "
    #                 f"member of another group ({order.get_group(kept_value)})"
    #             )

    #             # replacing group leader
    #             order.replace_group_leader(discarded_value, kept_value)

    #         # updating Carver values_orders and labels_per_values
    #         self.values_orders.update({feature: order})
    #         self.labels_per_values = self._get_labels_per_values(self.ordinal_encoding)


def transform_quantitative_feature(
    feature: BaseFeature, df_feature: Series, x_len: int
) -> tuple[str, list]:
    """Transforms a quantitative feature"""

    # keeping track of original index
    raw_index = df_feature.index

    # identifying nans
    feature_nans = (df_feature == feature.nan) | df_feature.isna()

    # converting nans to there corresponding quantile (if it was grouped to a quantile)
    if any(feature_nans):
        # quantile with which nans have been grouped
        nan_group = feature.values.get_group(feature.nan)

        # checking that nans have been grouped to a quantile
        if nan_group == feature.nan:
            nan_group = nan

        # converting to quantile value if grouped else keeping np.nan
        df_feature.mask(feature_nans, nan_group, inplace=True)

    # list of masks of values to replace with there respective group
    values_to_group = [df_feature <= value for value in feature.values if value != feature.nan]

    # corressponding group for each value
    group_labels = [
        [feature.label_per_value[value]] * x_len for value in feature.values if value != feature.nan
    ]

    # checking for values to group
    # if len(values_to_group) > 0:  # TODO check if this is needed
    df_feature = Series(select(values_to_group, group_labels, default=df_feature), index=raw_index)

    # reinstating nans otherwise nan is converted to 'nan' by numpy
    if any(feature_nans):
        df_feature[feature_nans] = feature.label_per_value.get(feature.nan, nan)

    return feature.version, list(df_feature)



""" Loads association filters."""

from .base_filters import BaseFilter, NonDefaultValidFilter, ValidFilter
from .qualitative_filters import CramervFilter, QualitativeFilter, TschuprowtFilter
from .quantitative_filters import PearsonFilter, QuantitativeFilter, SpearmanFilter

__all__ = [
    "BaseFilter",
    "QuantitativeFilter",
    "ValidFilter",
    "NonDefaultValidFilter",
    "SpearmanFilter",
    "PearsonFilter",
    "QualitativeFilter",
    "CramervFilter",
    "TschuprowtFilter",
]



""" defines a target rate"""

from abc import ABC, abstractmethod
from typing import Union

from pandas import DataFrame, Series


class TargetRate(ABC):
    """Target rate class."""

    __name__ = "target_rate"

    @abstractmethod
    def _compute(self, xagg: Union[DataFrame, Series]) -> Series:
        """Computes the target rate.

        Parameters
        ----------
        xagg : DataFrame
            A crosstab.

        Returns
        -------
        Series
            Target rate.
        """

    @abstractmethod
    def compute(self, xagg: Union[DataFrame, Series]) -> Series:
        """Computes the target rate.

        Parameters
        ----------
        xagg : DataFrame
            A crosstab.

        Returns
        -------
        Series
            Target rate.
        """



""" Loads Features tools."""

from .features import Features, get_names, get_versions
from .qualitatives import (
    CategoricalFeature,
    OrdinalFeature,
    QualitativeFeature,
    get_categorical_features,
    get_ordinal_features,
    get_qualitative_features,
)
from .quantitatives import QuantitativeFeature, get_quantitative_features
from .utils.base_feature import BaseFeature
from .utils.grouped_list import GroupedList

__all__ = [
    "Features",
    "get_names",
    "get_versions",
    "get_quantitative_features",
    "get_qualitative_features",
    "get_categorical_features",
    "get_ordinal_features",
    "BaseFeature",
    "GroupedList",
    "CategoricalFeature",
    "OrdinalFeature",
    "QuantitativeFeature",
    "QualitativeFeature",
]



""" Set of combination tools for Carvers"""

from .binary.binary_combination_evaluators import (
    BinaryCombinationEvaluator,
    CramervCombinations,
    TschuprowtCombinations,
)
from .continuous.continuous_combination_evaluators import (
    ContinuousCombinationEvaluator,
    KruskalCombinations,
)
from .utils.combination_evaluator import CombinationEvaluator

__all__ = [
    "CombinationEvaluator",
    "BinaryCombinationEvaluator",
    "TschuprowtCombinations",
    "CramervCombinations",
    "ContinuousCombinationEvaluator",
    "KruskalCombinations",
]



"""Tools to build simple buckets out of Quantitative and Qualitative features
for a binary classification model.
"""

from pandas import DataFrame, Series

from ..features import Features
from ..utils import extend_docstring
from .qualitatives import QualitativeDiscretizer
from .quantitatives import QuantitativeDiscretizer
from .utils.base_discretizer import BaseDiscretizer, Sample


class Discretizer(BaseDiscretizer):
    """Automatic discretization pipeline of continuous, discrete, categorical and ordinal features.

    Pipeline steps: :ref:`QuantitativeDiscretizer`, :ref:`QualitativeDiscretizer`.

    Modalities/values of features are grouped according to there respective orders:

    * [Categorical features] order based on modality target rate.
    * [Ordinal features] user-specified order.
    * [Continuous/Discrete features] real order of the values.
    """

    __name__ = "Discretizer"

    @extend_docstring(BaseDiscretizer.__init__)
    def __init__(
        self,
        features: Features,
        min_freq: float,
        **kwargs,
    ) -> None:
        # Initiating BaseDiscretizer
        super().__init__(features, **dict(kwargs, min_freq=min_freq))

    @extend_docstring(BaseDiscretizer.fit)
    def fit(self, X: DataFrame, y: Series) -> None:  # pylint: disable=W0222
        # Checking for binary target and copying X
        sample = self._prepare_data(Sample(X, y))

        # fitting quantitative features if any
        self._fit_quantitatives(**sample)

        # fitting qualitative features if any
        self._fit_qualitatives(**sample)

        # discretizing features based on each feature's values_order
        super().fit(**sample)

        return self

    def _fit_qualitatives(self, X: DataFrame, y: Series) -> None:
        """Fit the QualitativeDiscretizer on the qualitative features."""

        # [Qualitative features] Grouping qualitative features
        if len(self.features.qualitatives) > 0:
            # grouping qualitative features
            qualitative_discretizer = QualitativeDiscretizer(
                qualitatives=self.features.qualitatives, **dict(self.kwargs, copy=False)
            )
            qualitative_discretizer.fit(X, y)

    def _fit_quantitatives(self, X: DataFrame, y: Series) -> None:
        """Fit the QuantitativeDiscretizer on the quantitative features."""

        # [Quantitative features] Grouping quantitative features
        if len(self.features.quantitatives) > 0:
            # grouping quantitative features
            quantitative_discretizer = QuantitativeDiscretizer(
                quantitatives=self.features.quantitatives, **dict(self.kwargs, copy=False)
            )
            quantitative_discretizer.fit(X, y)



""" set of qualitative features"""

from .categorical_feature import CategoricalFeature, get_categorical_features
from .ordinal_feature import OrdinalFeature, get_ordinal_features
from .qualitative_feature import QualitativeFeature, get_qualitative_features, nan_unique

__all__ = [
    "OrdinalFeature",
    "CategoricalFeature",
    "QualitativeFeature",
    "get_ordinal_features",
    "get_categorical_features",
    "get_qualitative_features",
    "nan_unique",
]



""" TODO: initiate features from dataset
"""

from abc import ABC, abstractmethod
from typing import Any, Union

from pandas import DataFrame, Series, option_context

from ...config import Constants
from .grouped_list import GroupedList
from .serialization import json_deserialize_content, json_serialize_feature


class BaseFeature(ABC):
    """Base class for all features

    Parameters
    ----------
    name : str

    Attributes
    ----------
    name : str
        Name of the feature.
    has_nan : bool
        Whether or not feature has some NaNs.
    nan : any
        Value used to represent NaNs.
    has_default : bool
        Whether or not feature has some default values.
    default : any
        Default value.
    dropna : bool
        Whether or not nans must be removed.
    is_fitted : bool
        Whether or not feature has been fitted.
    ordinal_encoding : bool
        Whether or not to ordinally encode labels.
    values : GroupedList
        Feature values.
    labels : GroupedList
        Labels associated to feature's values.
    label_per_value : dict[str, str]
        Label for each value.
    value_per_label : dict[str, str]
        Value for each label.
    is_ordinal : bool
        Whether or not feature is ordinal.
    is_categorical : bool
        Whether or not feature is categorical.
    is_qualitative : bool
        Whether or not feature is qualitative.
    is_quantitative : bool
        Whether or not feature is quantitative.
    max_n_chars : int
        Max number of characters per label.
    statistics : dict[str, Any]
        Feature's trained statistics.
    history : list[dict, Any]
        Feature's combination history.
    raw_order : list[str]
        Base ordering for quantitative features.
    version : str
        Feature version.
    version_tag : str
        Feature version tag.
    """

    __name__ = "Feature"

    is_quantitative = False
    """Whether or not feature is quantitative"""

    is_qualitative = False
    """Whether or not feature is qualitative"""

    is_categorical = False
    """Whether or not feature is categorical"""

    is_ordinal = False
    """Whether or not feature is ordinal"""

    def __init__(self, name: str, **kwargs) -> None:
        """
        Parameters
        ----------
        name : str
            Name of the feature

        Keyword Arguments
        -----------------
        ordinal_encoding : bool, optional
            Whether or not to ordinal encode labels, by default ``False``
        nan : str, optional
            Label for missing values, by default ``"__NAN__"``
        default : str, optional
            Label for default values, by default ``"__OTHER__"``
        """
        self.name = name

        # whether or not feature has some NaNs
        self._has_nan = kwargs.get("has_nan", False)
        self.nan = kwargs.get("nan", Constants.NAN)

        # whether or not feature has some default values
        self._has_default = kwargs.get("has_default", False)
        self.default = kwargs.get("default", Constants.DEFAULT)

        # whether or not nans must be removed
        self._dropna = kwargs.get("dropna", False)

        # whether or not feature has been fitted
        self.is_fitted = kwargs.get("is_fitted", False)

        # whether or not to ordinally encode labels
        self._ordinal_encoding = kwargs.get("ordinal_encoding", False)

        # feature values, type and labels
        self.values = None  # current values
        self._labels = None  # current labels
        self.label_per_value: dict[str, str] = {}  # current label for all existing values
        self.value_per_label: dict[str, str] = {}  # label for all existing values

        # initating feature dtypes
        self.is_ordinal = kwargs.get("is_ordinal", self.is_ordinal)
        self.is_categorical = kwargs.get("is_categorical", self.is_categorical)
        self.is_qualitative = kwargs.get("is_qualitative", self.is_qualitative)
        self.is_quantitative = kwargs.get("is_quantitative", self.is_quantitative)

        # max number of characters per label
        self.max_n_chars = kwargs.get("max_n_chars", 50)

        # initiating feature's trained statistics
        self._statistics: dict[str:Any] = kwargs.get("statistics", None)

        # measures and filters used by selectors
        self.measures = kwargs.get("measures", {})
        self.filters = kwargs.get("filters", {})

        # initiating feature's combination history
        self._history: list[dict] = kwargs.get("history", None)

        # initiating base ordering (quantitative features)
        self.raw_order: list[str] = kwargs.get("raw_order", [])

        # initiating feature version
        self.version: str = kwargs.get("version", self.name)
        self.version_tag: str = kwargs.get("version_tag", self.name)

    def __repr__(self) -> str:
        """Feature representation"""
        return f"{self.__name__}('{self.version}')"

    @property
    def has_nan(self) -> bool:
        """Wether or not feature has nans"""
        return self._has_nan

    @has_nan.setter
    def has_nan(self, value: bool) -> None:
        """sets has_nan attribute"""
        if not isinstance(value, bool):
            raise ValueError(f"Trying to set has_nan with type {type(value)}")
        self._has_nan = value

    @property
    def statistics(self) -> DataFrame:
        """Feature's trained statistics"""
        # conversion to dataframe
        stats = self._statistics
        if stats is not None:
            stats = DataFrame(stats)

        # convertion to ordinal encoding if requested
        if self.ordinal_encoding and stats is not None:
            stats_copy = stats.copy()
            rev_value_per_label = {v: k for k, v in self.value_per_label.items()}

            # adding nan to the dictionary
            rev_value_per_label[self.nan] = self.label_per_value.get(self.nan)
            stats_copy.index = list(map(rev_value_per_label.get, stats_copy.index))
            return stats_copy
        return stats

    @statistics.setter
    def statistics(self, value: DataFrame) -> None:
        """Feature's trained statistics"""

        # case for binary targets
        if isinstance(value, DataFrame):
            self._statistics = value.to_dict()

        # case for continuous targets
        elif isinstance(value, Series):
            self._statistics = value.to_frame().to_dict()

        # case for selectors
        elif isinstance(value, dict):
            if self._statistics is None:
                self._statistics = {}
            self._statistics.update(value)
        else:
            raise ValueError(f"Trying to set statistics with type {type(value)}")

    @property
    def history(self) -> DataFrame:
        """Feature's combination history"""
        if self._history is not None:
            return DataFrame(self._history)
        return []

    @history.setter
    def history(self, value: list[dict]) -> None:
        """Feature's combination history"""
        # initiating history
        if self._history is None:
            self._history: list[dict] = []
        self._history += value

    @property
    def has_default(self) -> bool:
        """Whether or not the feature has default values"""
        return self._has_default

    @has_default.setter
    def has_default(self, value: bool) -> None:
        """sets ordinal_encoding and labels accordingly"""

        # check input value
        if not isinstance(value, bool):
            raise ValueError(f"Trying to set has_default with type {type(value)}")

        # copying values
        values = GroupedList(self.values)

        # setting has_default
        if value and not self.has_default:
            # adding to the values
            values.append(self.default)

            # updating labels
            self.update(values, replace=True)

        # checking that it was not already set to True
        elif not value and self.has_default:
            raise RuntimeError(f"[{self}] has_default has been set to True, can't go back")

        # updating attribute
        self._has_default = value

    @property
    def ordinal_encoding(self) -> bool:
        """whether or not to ordinally encode feature's labels"""
        return self._ordinal_encoding

    @ordinal_encoding.setter
    def ordinal_encoding(self, value: bool) -> None:
        """sets ordinal_encoding and labels accordingly"""

        # check input value
        if not isinstance(value, bool):
            raise ValueError(f"Trying to set ordinal_encoding with type {type(value)}")

        # updating attribute
        self._ordinal_encoding = value

        # updating labels
        if self.values is not None:
            self.update_labels()

    @property
    def dropna(self) -> bool:
        """whether or not to drop missing values (nan)"""
        return self._dropna

    @dropna.setter
    def dropna(self, value: bool) -> None:
        """Activates or deactivates feature's dropna mode"""

        # check input value
        if not isinstance(value, bool):
            raise ValueError(f"Trying to set dropna with type {type(value)}")

        # check input value
        if self.values is None:
            raise ValueError("Trying to set dropna before there were values observed")

        # activating dropna mode
        if value and not self.dropna:
            # adding nan to the values only if they were found
            if self.has_nan and not self.values.contains(self.nan):
                values = GroupedList(self.values)
                values.append(self.nan)

                # updating values
                self.update(values, replace=True)

        # deactivating dropna mode
        elif not value and self.dropna:
            # checking for values merged with nans
            if self.values is not None and len(self.values.get(self.nan)) > 1:
                raise RuntimeError(
                    "Can not set feature dropna=False has values were grouped with nans."
                )

            # dropping nans from values
            values = GroupedList(self.values)
            if self.nan in self.values:
                values.remove(self.nan)

            # updating values
            self.update(values, replace=True)

        # setting dropna
        self._dropna = value

    @property
    def content(self) -> dict:
        """returns feature values' content"""
        if isinstance(self.values, GroupedList):
            return self.values.content
        return self.values

    @property
    def labels(self) -> GroupedList:
        """gives labels associated to feature's values"""
        # default labels are values
        return self._labels

    @labels.setter
    def labels(self, raw_labels: GroupedList) -> None:
        """updates labels per values and values per label associated to feature's labels"""

        # requested float output (AutoCarver) -> converting to integers
        labels = raw_labels[:]
        if self.ordinal_encoding:
            labels = [n for n, _ in enumerate(labels)]

        # updating list of labels
        self._labels = list(labels)

        # updating label_per_value accordingly
        self._update_value_per_label(raw_labels=raw_labels)

    def _update_value_per_label(self, raw_labels: list[str]) -> None:
        """updates value per label and label per value"""

        # iterating over values and labels
        self.value_per_label = {}
        for value, label, raw_label in zip(self.values, self._labels, raw_labels):
            # updating label_per_value
            for grouped_value in self.values.get(value):
                self.label_per_value.update({grouped_value: label})

            # getting value
            value_to_keep = value
            if self.ordinal_encoding:
                value_to_keep = raw_label

            # updating value_per_label
            self.value_per_label.update({label: value_to_keep})

    @abstractmethod
    def make_labels(self) -> GroupedList:
        """builds labels according to feature's values"""
        # default labels are values
        return self.values

    @abstractmethod
    def _make_summary(self) -> dict:
        """returns a summary of the feature"""

    @property
    def summary(self) -> dict:
        """Summary of feature's discretization process"""
        return self._make_summary()

    def _add_statistics_to_summary(self, summary: list[dict]) -> list[dict]:
        """adds statistics to summary"""
        # adding statistics
        if self.statistics is not None:
            # iterating over each modality
            for label_content in summary:
                label = label_content["label"]
                statistics = self.statistics.loc[label].to_dict()
                label_content.update(statistics)

        # adding hisory
        history = self.history
        if len(history) > 0:
            selected = {}

            # checking for viable combination without dropna
            with option_context("future.no_silent_downcasting", True):
                viable = history["viable"].fillna(False).astype(bool)
            if viable.any():
                # checking for requested dropna
                dropna = history["dropna"].fillna(False)
                if dropna.any():
                    # checking for viable combination with dropna
                    if history[dropna].viable.any():
                        selected = history[viable & dropna].iloc[0].to_dict()

                # best_combination for without dropna
                else:
                    selected = history[viable].iloc[0].to_dict()

            # TODO checking for match between selected combination and summary
            # removing unwanted keys
            selected.pop("viable", None)
            selected.pop("dropna", None)
            selected.pop("combination", None)
            selected.pop("info", None)
            selected.pop("train", None)
            selected.pop("dev", None)

            # adding selected combination to summary
            for label_content in summary:
                label_content.update(selected)

        return summary

    def update_labels(self) -> None:
        """updates label for each value of the feature"""

        # initiating labels for qualitative features
        raw_labels = self.make_labels()

        # saving updated labels
        self.labels = raw_labels

    def _update_statistics_value(
        self, kept_label: Union[str, float], kept_value: Union[str, float]
    ) -> None:
        """updates feature's statistics index with values"""

        # updating feature's statistics
        if self.statistics is not None and isinstance(self.statistics, DataFrame):
            self.statistics.rename(index={kept_label: kept_value}, inplace=True)

    @abstractmethod
    def _specific_update(self, values: GroupedList, convert_labels: bool = False) -> None:
        """update content of values specifically per feature type"""

    def update(
        self,
        values: GroupedList,
        convert_labels: bool = False,
        sorted_values: bool = False,
        replace: bool = False,
    ) -> None:
        """updates content of values of the feature"""

        # values are the same but sorted
        if sorted_values:
            self.values = self.values.sort_by(values)

        # checking for GroupedList
        elif not isinstance(values, GroupedList):
            raise ValueError(f"[{self}] Wrong input, expected GroupedList object.")

        # replacing existing values
        elif replace:
            self.values = values

        # using type method specific
        else:
            self._specific_update(values, convert_labels=convert_labels)

        # updating labels accordingly
        self.update_labels()

    def fit(self, X: DataFrame, y: Series = None) -> None:
        """Fits the feature to a DataFrame"""
        _, _ = X, y  # unused attributes

        # checking for previous fit
        if self.is_fitted:
            raise RuntimeError(f"[{self}] Already been fitted!")

        # looking for NANS
        if any(X[self.name].isna()):
            self.has_nan = True

        self.is_fitted = True  # feature is fitted

    def check_values(self, X: DataFrame) -> None:
        """checks for unexpected values from unique values in DataFrame"""

        # checking for nans whereas at training none were witnessed
        if (any(X[self.version].isna()) or any(X[self.version] == self.nan)) and not self.has_nan:
            raise ValueError(f"[{self}] Unexpected NaNs.")

    def group(self, to_discard: list[str], to_keep: str) -> None:
        """wrapper of GroupedList: groups a list of values into a kept value"""

        # using GroupedList's group
        values = GroupedList(self.values)
        values.group(to_discard, to_keep)

        # updating feature's values
        self.update(values, replace=True)

    def _historize(self, combination: dict) -> None:
        """historizes a combination"""
        self.history.append(combination)

    def to_json(self, light_mode: bool = False) -> dict:
        """Converts to JSON format.

        To be used with ``json.dump``.

        Parameters
        ----------
        light_mode: bool, optional
            Whether or not to save feature's history and statistics, by default False

        Returns
        -------
        str
            JSON serialized object
        """

        # minimal output json
        feature = {
            "name": self.name,
            "version": self.version,
            "version_tag": self.version_tag,
            "has_nan": self.has_nan,
            "nan": self.nan,
            "has_default": self.has_default,
            "default": self.default,
            "dropna": self.dropna,
            "is_fitted": self.is_fitted,
            "values": self.values,
            "content": self.content,
            "is_ordinal": self.is_ordinal,
            "is_categorical": self.is_categorical,
            "is_qualitative": self.is_qualitative,
            "is_quantitative": self.is_quantitative,
            "raw_order": self.raw_order,
            "ordinal_encoding": self.ordinal_encoding,
            "statistics": self._statistics,
            "measures": self.measures,
            "filters": self.filters,
        }

        # enriched mode (larger json)
        if not light_mode:
            feature.update({"history": self._history})

        return json_serialize_feature(feature)

    @classmethod
    def load(cls, feature_json: dict) -> "BaseFeature":
        """Loads a feature"""

        # deserializing content into grouped list of values
        values = json_deserialize_content(feature_json)

        # initiating feature without content
        feature = cls(**dict(feature_json, load_mode=True))

        # updating feature with deserialized content
        if values is not None:
            feature.update(values, replace=True)

        return feature



""" Loads Discretization base tools."""

from .base_discretizer import BaseDiscretizer, Sample
from .type_discretizers import StringDiscretizer

__all__ = ["BaseDiscretizer", "StringDiscretizer", "Sample"]



""" set of tests for combinations within carvers"""

from enum import Enum

from numpy import isclose
from pandas import DataFrame, Series


class TestKeys(str, Enum):
    """keys for test results"""

    RANKS_TRAIN_DEV = "ranks_train_dev"
    MIN_FREQ = "min_freq"
    DISTINCT_RATES = "distinct_rates"
    VIABLE = "viable"
    INFO = "info"


class TestMessages(str, Enum):
    """messages for test results"""

    INVERSION_RATES = "Inversion of target rates per modality"
    NON_REPRESENTATIVE = "Non-representative modality for min_freq={min_freq:2.2%}"
    NON_DISTINCT_RATES = "Non-distinct target rates per consecutive modalities"
    PASSED_TESTS = ""


def _test_distinct_target_rates_between_modalities(target_rates: Series) -> bool:
    """tests for distinct target rates between consecutive modalities"""

    # - target rates are distinct for consecutive modalities
    return not any(isclose(target_rates[1:], target_rates.shift(1)[1:]))


def _test_minimum_frequency_per_modality(frequencies: Series, min_freq: float) -> bool:
    """tests that minimum frequency has been reached for each modality"""
    # - minimum frequency is reached for all modalities
    return all(frequencies >= min_freq)


def _test_modality_ordering(train_target_rate: Series, dev_target_rate: Series) -> bool:
    """tests whether train and dev targets rates are similarly ranked between datasets"""
    # - grouped values have the same ranks in train/test
    return all(
        train_target_rate.sort_index().sort_values().index
        == dev_target_rate.sort_index().sort_values().index
    )


def is_viable(test_results: dict):
    """checks if combination is viable on train and dev (if provided)"""

    return test_results["train"][TestKeys.VIABLE.value] and (
        test_results["dev"][TestKeys.VIABLE.value]
        or test_results["dev"][TestKeys.VIABLE.value] is None
    )


def test_viability(
    rates: DataFrame, min_freq: float, target_rate: str, train_target_rate: Series = None
) -> dict:
    """tests viability of the rates"""

    # - minimum frequency is reached for all modalities
    min_freq_test = _test_minimum_frequency_per_modality(rates["frequency"], min_freq)

    # - target rates are distinct for all modalities
    distinct_rates = _test_distinct_target_rates_between_modalities(rates[target_rate])

    # gathering results
    test_results = {
        TestKeys.VIABLE.value: min_freq_test and distinct_rates,
        TestKeys.MIN_FREQ.value: min_freq_test,
        TestKeys.DISTINCT_RATES.value: distinct_rates,
    }

    # adding ranking test if train_rates where provided
    if train_target_rate is not None:
        ordering = _test_modality_ordering(train_target_rate, rates[target_rate])
        test_results.update(
            {
                TestKeys.RANKS_TRAIN_DEV.value: ordering,
                TestKeys.VIABLE.value: test_results[TestKeys.VIABLE.value] and ordering,
            }
        )

        # return tests on dev
        return {"dev": add_info(test_results, min_freq)}

    # return tests on train
    return {
        "train": add_info(test_results, min_freq),
        "train_rates": rates,
        TestKeys.VIABLE.value: test_results[TestKeys.VIABLE.value],
    }


def add_info(test_results: dict[str, bool], min_freq: float) -> dict[str, str]:
    """Adds information to test results."""

    messages: list[str] = []
    if not test_results.get(TestKeys.RANKS_TRAIN_DEV.value, True):
        messages.append(TestMessages.INVERSION_RATES.value)
    if not test_results.get(TestKeys.MIN_FREQ.value, True):
        messages.append(TestMessages.NON_REPRESENTATIVE.value.format(min_freq=min_freq))
    if not test_results.get(TestKeys.DISTINCT_RATES.value, True):
        messages.append(TestMessages.NON_DISTINCT_RATES.value)

    if not messages:
        messages.append(TestMessages.PASSED_TESTS.value)

    return {
        TestKeys.VIABLE.value: test_results[TestKeys.VIABLE.value],
        TestKeys.INFO.value: "; ".join(messages),
    }



""" Loads feature selection tools."""

# from .base_selector import BaseSelector
from .classification_selector import ClassificationSelector
from .filters import (
    BaseFilter,
    CramervFilter,
    PearsonFilter,
    QualitativeFilter,
    QuantitativeFilter,
    SpearmanFilter,
    TschuprowtFilter,
)
from .measures import (
    AbsoluteMeasure,
    BaseMeasure,
    Chi2Measure,
    CramervMeasure,
    DistanceMeasure,
    IqrOutlierMeasure,
    KruskalMeasure,
    ModeMeasure,
    NanMeasure,
    OutlierMeasure,
    PearsonMeasure,
    RMeasure,
    SpearmanMeasure,
    TschuprowtMeasure,
    ZscoreOutlierMeasure,
)
from .regression_selector import RegressionSelector

__all__ = [
    # selectors
    "RegressionSelector",
    "ClassificationSelector",
    # filters
    "BaseFilter",
    "QuantitativeFilter",
    "SpearmanFilter",
    "PearsonFilter",
    "QualitativeFilter",
    "CramervFilter",
    "TschuprowtFilter",
    # measures
    "AbsoluteMeasure",
    "ModeMeasure",
    "NanMeasure",
    "BaseMeasure",
    "OutlierMeasure",
    "Chi2Measure",
    "CramervMeasure",
    "PearsonMeasure",
    "SpearmanMeasure",
    "TschuprowtMeasure",
    "RMeasure",
    "DistanceMeasure",
    "IqrOutlierMeasure",
    "KruskalMeasure",
    "ZscoreOutlierMeasure",
]



""" Constants used throughout the modules. """

from dataclasses import dataclass
from typing import Final


@dataclass
class Constants:
    """Constants used throughout the modules."""

    DEFAULT: Final[str] = "__OTHER__"
    NAN: Final[str] = "__NAN__"
    INF: Final[str] = "__INF__"



"""Tools to build simple buckets out of Qualitative features
for a binary classification model.
"""

from numpy import arange, argmin, array, nan_to_num, vstack
from pandas import DataFrame, Series, notna

from ...features import GroupedList, OrdinalFeature
from ...utils import extend_docstring
from ..utils.base_discretizer import BaseDiscretizer, Sample


class OrdinalDiscretizer(BaseDiscretizer):
    """Automatic discretization of ordinal features, grouping less frequent modalities with the
    closest modlity in target rate or by frequency.

    NaNs are left untouched.

    Only use for qualitative ordinal features.

    Fisrt fits :ref:`StringDiscretizer` if neccesary.
    """

    __name__ = "OrdinalDiscretizer"

    @extend_docstring(BaseDiscretizer.__init__, append=False, exclude=["features"])
    def __init__(
        self,
        ordinals: list[OrdinalFeature],
        min_freq: float,
        **kwargs,
    ):
        """
        Parameters
        ----------
        ordinals : list[OrdinalFeature]
            Ordinal features to process
        """
        # Initiating BaseDiscretizer
        super().__init__(ordinals, **dict(kwargs, min_freq=min_freq))

    def _prepare_data(self, sample: Sample) -> Sample:  # pylint: disable=W0222
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in
            ``OrdinalDiscretizer.features``.

        y : Series
            Binary target feature.

        Returns
        -------
        DataFrame
            A formatted copy of X
        """
        # checking for binary target and copying X
        sample = super()._prepare_data(sample)

        # fitting features
        self.features.fit(**sample)

        # filling up nans for features that have some
        sample.X = self.features.fillna(sample.X)

        return sample

    @extend_docstring(BaseDiscretizer.fit)
    def fit(self, X: DataFrame, y: Series) -> None:  # pylint: disable=W0222
        # checking values orders
        sample = self._prepare_data(Sample(X, y))

        # verbose if requested
        self._log_if_verbose()

        # grouping rare modalities for each feature
        common_modalities = {
            feature.version: find_common_modalities(
                sample.X[feature.version],
                sample.y,
                min_freq=self.min_freq,
                labels=[label for label in feature.labels if label != feature.nan],
            )
            for feature in self.features
        }

        # updating features accordingly
        self.features.update(common_modalities, convert_labels=True)

        # discretizing features based on each feature's values_order
        super().fit(**sample)

        return self


def find_common_modalities(
    df_feature: Series, y: Series, min_freq: float, labels: list[str]
) -> GroupedList:
    """finds common modalities of a ordinal feature"""

    # converting to grouped list
    labels = GroupedList(labels)

    # computing frequencies and target rate of each modality
    stats, len_df = compute_stats(df_feature, y, labels)

    # case 1: there are underrepresented modalities/values
    while any(stats[0, :] / len_df < min_freq) & (stats.shape[1] > 1):
        # identifying the first underrepresented value
        discarded_idx = argmin(stats[0, :])

        # choosing amongst previous and next modality (by volume and target rate)
        kept_idx = find_closest_modality(
            discarded_idx,
            stats[0, :] / len_df,
            stats[1, :] / stats[0, :],
            min_freq,
        )

        # grouping discarded idx with kept idx
        labels.group(labels[discarded_idx], labels[kept_idx])

        # updating stats accordingly
        stats = update_stats(stats, discarded_idx, kept_idx)

    # case 2 : no underrepresented value
    return labels


def update_stats(stats: array, discarded_idx: int, kept_idx: int) -> array:
    """Updates frequencies and target rates after grouping two modalities"""

    # adding up grouped frequencies and target counts
    stats[:, kept_idx] += nan_to_num(stats[:, discarded_idx], nan=0)

    # removing discarded modality
    return stats[:, arange(stats.shape[1]) != discarded_idx]


def compute_stats(df_feature: Series, y: Series, labels: GroupedList) -> tuple[array, int]:
    """Computes frequencies and target rates of each modality"""

    # filtering nans
    not_nans = notna(df_feature)

    # total size
    len_df = len(df_feature)

    # computing frequencies and target rates
    stats = vstack(
        (
            # frequencies
            df_feature[not_nans]
            .value_counts(dropna=False, normalize=False)
            .reindex(labels, fill_value=0)
            .values,
            # target rates
            y[not_nans].groupby(df_feature[not_nans]).sum().reindex(labels).values,
        )
    )

    return stats, len_df


def find_closest_modality(
    idx: int, frequencies: array, target_rates: array, min_freq: float
) -> int:
    """Finds the closest modality in terms of frequency and target rate"""

    # case 0: only one modality
    if frequencies.shape[0] == 1:
        return 0

    # case 1: lowest ranked modality
    if idx == 0:
        return 1

    # case 2: highest ranked modality
    if idx == frequencies.shape[0] - 1:
        return idx - 1

    # checking if next modality is closer
    if is_next_modality_closer(idx, frequencies, target_rates, min_freq):
        return idx + 1

    # by default closest modality is the previous one
    return idx - 1


def is_next_modality_closer(
    idx: int, frequencies: array, target_rates: array, min_freq: float
) -> bool:
    """Determines if the next modality is closer than the previous to the current one"""

    # Extract relevant frequencies and target rates
    previous_freq, current_freq, next_freq = frequencies[idx - 1 : idx + 2]

    # comparing frequencies to min_freq
    both_below_min_freq = (next_freq < min_freq) and (previous_freq < min_freq)
    both_above_min_freq = (next_freq >= min_freq) and (previous_freq >= min_freq)

    # case 1: no observation to differentiate by target rate -> least frequent modality is choosen
    if current_freq == 0:
        return next_freq < previous_freq

    # case 2: next modality is the only below min_freq -> underrepresented modality is choosen
    if next_freq < min_freq <= previous_freq:
        return True

    # case 3: both are below or above min_freq -> closest modality by target rate
    if both_below_min_freq or both_above_min_freq:
        return is_next_modality_closer_by_target_rate(idx, target_rates)

    # by default closest modality is the previous one
    return False


def is_next_modality_closer_by_target_rate(idx: int, target_rates: array) -> bool:
    """Determines if the next modality is closer in terms of target rate than the previous to the
    current one"""

    # Extract relevant frequencies and target rates
    previous_target, current_target, next_target = target_rates[idx - 1 : idx + 2]

    # absolute difference is less for the next modality
    if abs(previous_target - current_target) > abs(next_target - current_target):
        return True

    # absolute difference is less for the previous modality
    return False



"""Tool to build optimized buckets out of Quantitative and Qualitative features
for continuous regression tasks.
"""

from numpy import unique
from pandas import DataFrame, Series

from ..combinations import KruskalCombinations
from ..features import BaseFeature, Features
from ..utils import extend_docstring
from .utils.base_carver import BaseCarver, Samples


class ContinuousCarver(BaseCarver):
    """Automatic carving of continuous, discrete, categorical and ordinal
    features that maximizes association with a continuous target.

    For continuous targets, :ref:`Kruskal` is used as association measure to sort combinations.

    Examples
    --------
    `Continuous Regression Example <https://autocarver.readthedocs.io/en/latest/examples/Carvers/
    ContinuousRegression/continuous_regression_example.html>`_
    """

    __name__ = "ContinuousCarver"
    is_y_continuous = True

    @extend_docstring(BaseCarver.__init__)
    def __init__(
        self,
        features: Features,
        min_freq: float,
        dropna: bool = True,
        max_n_mod: int = 5,
        **kwargs,
    ) -> None:
        """
        Keyword Arguments
        -----------------
        combinations : ContinuousCombinationEvaluator, optional
            Metric to perform association measure between :class:`Features` and target.

            Currently, only :ref:`KruskalCombinations` are implemented.
        """
        # default binary combinations
        combinations = kwargs.pop("combinations", None)
        if combinations is None:
            combinations = KruskalCombinations(max_n_mod=max_n_mod)

        # association measure used to find the best groups for binary targets
        if not combinations.is_y_continuous:
            raise ValueError(
                f"[{self.__name__}] {combinations} is not suited for continuous targets. "
                f"Choose from: KruskalCombinations."
            )

        # Initiating BaseCarver
        super().__init__(
            features=features,
            min_freq=min_freq,
            combinations=combinations,
            dropna=dropna,
            **kwargs,
        )

    def _prepare_data(self, samples: Samples) -> Samples:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in
            ``AutoCarver.features``.

        y : Series
            Binary target feature with wich the association is maximized.

        X_dev : DataFrame, optional
            Dataset to evalute the robustness of discretization, by default ``None``
            It should have the same distribution as X.

        y_dev : Series, optional
            Binary target feature with wich the robustness of discretization is evaluated,
            by default ``None``

        Returns
        -------
        tuple[DataFrame, DataFrame]
            Copies of (X, X_dev) to be used according to target type
        """

        # continuous target, checking values
        if str in samples.train.y.apply(type).unique():
            raise ValueError(
                f"[{self.__name__}] y must be a continuous Series (int or float, not object)"
            )

        y_values = unique(samples.train.y)
        if len(y_values) <= 2:
            raise ValueError(
                f"[{self.__name__}] provided y is binary, consider using BinaryCarver instead."
            )
        # Checking for binary target and copying X
        return super()._prepare_data(samples)

    def _aggregator(self, X: DataFrame, y: Series) -> dict[str, DataFrame]:
        """Computes y values for modalities of specified features and ensures the ordering
        according to the known labels

        Parameters
        ----------
        features : list[str]
            _description_
        X : DataFrame
            _description_
        y : Series
            _description_
        labels_orders : dict[str, GroupedList]
            _description_

        Returns
        -------
        dict[str, DataFrame]
            _description_
        """
        # checking for empty datasets
        yvals = {feature.version: None for feature in self.features}
        if X is not None:
            # list of y values for each modality of X
            for feature in self.features:
                yvals.update({feature.version: get_target_values_by_modality(X, y, feature)})

        return yvals


def get_target_values_by_modality(X: DataFrame, y: Series, feature: BaseFeature) -> dict:
    """Computes y values for modalities of specified features and ensures the ordering
    according to the known labels"""

    # list of y values for each modality of X
    yval = y.groupby(X[feature.version]).apply(lambda u: list(u))

    # reindexing to ensure the right order
    return yval.reindex(feature.labels, fill_value=[])



"""Tool to build optimized buckets out of Quantitative and Qualitative features
for multiclass classification tasks.
"""

from typing import Any

from pandas import DataFrame, Series, unique

from ..discretizers.utils.base_discretizer import BaseDiscretizer, Sample
from ..features import Features
from ..utils import extend_docstring
from .binary_carver import BinaryCarver
from .utils.base_carver import Samples


class MulticlassCarver(BinaryCarver):
    """Automatic carving of continuous, discrete, categorical and ordinal
    features that maximizes association with a multiclass target.


    Examples
    --------
    `Multiclass Classification Example <https://autocarver.readthedocs.io/en/latest/examples/
    Carvers/MulticlassClassification/multiclass_classification_example.html>`_
    """

    __name__ = "MulticlassCarver"
    is_y_binary = False
    is_y_multiclass = True

    @extend_docstring(BinaryCarver.__init__)
    def __init__(
        self,
        features: Features,
        min_freq: float,
        dropna: bool = True,
        max_n_mod: int = 5,
        **kwargs,
    ) -> None:
        """ """
        # Initiating BinaryCarver
        super().__init__(
            features=features, min_freq=min_freq, dropna=dropna, max_n_mod=max_n_mod, **kwargs
        )

        # warning user for
        if "copy" in kwargs and kwargs["copy"] is True:
            print(
                "WARNING: can't set copy=True for MulticlassCarver (no inplace DataFrame.assign)."
            )

    def _prepare_data(self, samples: Samples) -> Samples:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in
            ``AutoCarver.features``.

        y : Series
            Binary target feature with wich the association is maximized.

        X_dev : DataFrame, optional
            Dataset to evalute the robustness of discretization, by default ``None``
            It should have the same distribution as X.

        y_dev : Series, optional
            Binary target feature with wich the robustness of discretization is evaluated,
            by default ``None``

        Returns
        -------
        tuple[DataFrame, DataFrame, dict[str, Callable]]
            Copies of (X, X_dev) and helpers to be used according to target type
        """
        # converting target to str
        samples.train.y = samples.train.y.astype(str)

        # multiclass target, checking values
        if len(unique(samples.train.y)) <= 2:
            raise ValueError(
                f"[{self.__name__}] provided y is binary, consider using BinaryCarver instead."
            )

        # checking for dev target's values
        if samples.dev.y is not None:
            # converting target to str
            samples.dev.y = samples.dev.y.astype(str)

            # check that classes of y are classes of y_dev
            unique_y_dev = samples.dev.y.unique()
            unique_y = samples.train.y.unique()
            missing_y = [mod_y for mod_y in unique_y if mod_y not in unique_y_dev]
            missing_y_dev = [mod_y_dev for mod_y_dev in unique_y_dev if mod_y_dev not in unique_y]
            if len(missing_y) > 0 or len(missing_y_dev) > 0:
                raise ValueError(
                    f"[{self.__name__}] Mismatched classes between y and y_dev"
                    f": train({missing_y_dev}), dev({missing_y})"
                )

        return samples

    @extend_docstring(BinaryCarver.fit)
    def fit(
        self,
        X: DataFrame,
        y: Series,
        *,
        X_dev: DataFrame = None,
        y_dev: Series = None,
    ) -> None:
        # initiating samples
        samples = Samples(train=Sample(X, y), dev=Sample(X_dev, y_dev))

        # preparing datasets and checking for wrong values
        samples = self._prepare_data(samples)

        # getting distinct y classes
        y_classes = sorted(list(samples.train.y.unique()))[1:]  # removing one of the classes

        # adding versionned features
        self.features.add_feature_versions(y_classes)

        # iterating over each class minus one
        for n, y_class in enumerate(y_classes):
            if self.verbose:  # verbose if requested
                print(
                    f"\n---------\n[{self.__name__}] Fit y={y_class} ({n+1}/{len(y_classes)})"
                    "\n------"
                )

            # identifying this y_class
            train_y_class = get_one_vs_rest(samples.train.y, y_class)
            dev_y_class = get_one_vs_rest(samples.dev.y, y_class)

            # features for specific group
            class_features = Features(self.features.get_version_group(y_class))

            # initiating BinaryCarver for y_class
            binary_carver = BinaryCarver(
                features=class_features,
                combinations=self.combinations,
                **dict(
                    self.kwargs,
                    copy=True,
                    min_freq=self.min_freq,
                ),  # copying x to keep raw columns as is
            )

            # fitting BinaryCarver for y_class
            binary_carver.fit_transform(
                samples.train.X, train_y_class, X_dev=samples.dev.X, y_dev=dev_y_class
            )

            # filtering out dropped features whilst keeping other version tags
            kept_features = binary_carver.features.versions
            kept_features += [
                feature.version for feature in self.features if feature.version_tag != y_class
            ]
            self.features.keep(kept_features)

            if self.verbose:  # verbose if requested
                print("---------\n")

        # initiating BaseDiscretizer with features for each y_class
        BaseDiscretizer.__init__(
            self,
            features=self.features,
            **dict(
                self.kwargs,
                min_freq=self.min_freq,
                dropna=self.dropna,
                combinations=self.combinations,
            ),
        )

        # fitting BaseDiscretizer
        BaseDiscretizer.fit(self, samples.train.X, samples.train.y)

        return self


def get_one_vs_rest(y: Series, y_class: Any) -> Series:
    """converts a multiclass target into binary of specific y_class"""
    if y is not None:
        return (y == y_class).astype(int)
    return None



"""Base tools to convert values into specific types.
"""

from pandas import DataFrame, Series

from ...features import BaseFeature, Features, GroupedList
from ...features.qualitatives import nan_unique
from ...utils import extend_docstring
from .base_discretizer import BaseDiscretizer, Sample
from .multiprocessing import apply_async_function


class StringDiscretizer(BaseDiscretizer):
    """Converts specified columns of a DataFrame into strings.
    First step of a Qualitative discretization pipeline.

    * Keeps NaN inplace
    * Converts floats of int to int
    """

    __name__ = "StringDiscretizer"

    @extend_docstring(BaseDiscretizer.__init__, exclude=["min_freq"])
    def __init__(self, features: Features, **kwargs) -> None:
        # initiating features
        features = Features(features, **kwargs)

        # Initiating BaseDiscretizer
        super().__init__(features=features, **kwargs)

    @extend_docstring(BaseDiscretizer.fit)
    def fit(self, X: DataFrame, y: Series = None) -> None:  # pylint: disable=W0222
        self._log_if_verbose()  # verbose if requested

        # checking for binary target and copying X
        sample = self._prepare_data(Sample(X, y))

        # transforming all features
        all_orders = apply_async_function(fit_feature, self.features, self.n_jobs, sample.X)

        # updating features accordingly
        self.features.update(dict(all_orders), replace=True)
        self.features.fit(**sample)

        # discretizing features based on each feature's values_order
        super().fit(X, y)

        return self


def fit_feature(feature: BaseFeature, df_feature: Series):
    """fits one feature"""

    # unique feature values
    unique_values = nan_unique(df_feature)

    # initiating feature values
    order = GroupedList(unique_values)
    if feature.values is not None:
        order = GroupedList(feature.values)

    # formatting values
    for value in unique_values:
        # case 0: the value is an integer
        if isinstance(value, float) and float.is_integer(value):
            str_value = str(int(value))  # converting value to string

        # case 1: converting to string
        else:
            str_value = str(value)

        # adding missing str_value to the order (for provided orders)
        if str_value not in order:
            order.append(str_value)

        # adding float/int values to the order (for provided orders)
        if value not in order:
            order.append(value)

        # grouping float/int value into the str value
        order.group(value, str_value)

    return feature.version, order


class TimedeltaDiscretizer(BaseDiscretizer):
    """TODO Converts specified columns of a DataFrame into float timedeltas.

    * Keeps NaN inplace
    """

    __name__ = "TimedeltaDiscretizer"

    @extend_docstring(BaseDiscretizer.__init__)
    def __init__(self, features: list[BaseFeature], **kwargs) -> None:
        """
        Parameters
        ----------
        features : list[str]
            List of column names of qualitative features to be converted as string
        """
        # initiating features
        features = Features(features, **kwargs)

        # Initiating BaseDiscretizer
        super().__init__(features=features, **kwargs)

    @extend_docstring(BaseDiscretizer.fit)
    def fit(self, X: DataFrame, y: Series = None) -> None:  # pylint: disable=W0222
        self._log_if_verbose()  # verbose if requested

        # checking for binary target and copying X
        sample = self._prepare_data(Sample(X, y))

        # transforming all features
        all_orders = apply_async_function(fit_feature, self.features, self.n_jobs, sample.X)

        # updating features accordingly
        self.features.update(dict(all_orders), replace=True)
        self.features.fit(**sample)

        # discretizing features based on each feature's values_order
        super().fit(X, y)

        return self



""" Loads Qualitative Discretization tools."""

from .categorical_discretizer import CategoricalDiscretizer
from .chained_discretizer import ChainedDiscretizer
from .ordinal_discretizer import OrdinalDiscretizer
from .qualitative_discretizer import QualitativeDiscretizer

__all__ = [
    "QualitativeDiscretizer",
    "CategoricalDiscretizer",
    "ChainedDiscretizer",
    "OrdinalDiscretizer",
]



""" Set of tools used for pretty printing"""

from pandas import DataFrame

from ...features import BaseFeature


def prettier_xagg(
    nice_xagg: DataFrame = None,
    caption: str = None,
    hide_index: bool = False,
) -> str:
    """Pretty display of frequency and target rate per modality on the same line

    Parameters
    ----------
    nice_xagg : DataFrame, optional
        Target rate and frequency per modality, by default None
    caption : str, optional
        Title of the HTML table, by default None
    hide_index : bool, optional
        Whether or not to hide the index (for dev distribution)

    Returns
    -------
    str
        HTML format of the crosstab
    """
    # checking for a provided xtab
    nicer_xagg = ""
    if nice_xagg is not None:
        # checking for non unique indices
        if any(nice_xagg.index.duplicated()):
            nice_xagg.reset_index(inplace=True)

        # adding coolwarm color gradient
        nicer_xagg = nice_xagg.style.background_gradient(cmap="coolwarm")

        # printing inline notebook
        nicer_xagg = nicer_xagg.set_table_attributes("style='display:inline'")

        # lower precision
        nicer_xagg = nicer_xagg.format(precision=4)

        # adding custom caption/title
        if caption is not None:
            nicer_xagg = nicer_xagg.set_caption(caption)

        # hiding index for dev
        if hide_index:
            nicer_xagg.hide(axis="index")

        # converting to html
        nicer_xagg = nicer_xagg._repr_html_()  # pylint: disable=W0212

    return nicer_xagg


def index_mapper(feature: BaseFeature, xtab: DataFrame = None) -> DataFrame:
    """Prints a binary xtab's statistics

    Parameters
    ----------
    order_get : Callable
        Ordering of modalities used to map indices
    xtab : Dataframe
        A crosstab, by default None

    Returns
    -------
    DataFrame
        Target rate and frequency per modality
    """
    # checking for an xtab
    mapped_xtab = None
    if xtab is not None:
        # copying initial xtab
        mapped_xtab = xtab.copy()

        # modifying indices based on provided order
        mapped_xtab.index = feature.labels[:]

    return mapped_xtab



""" Loads utils."""

from .attributes import get_attribute, get_bool_attribute
from .dependencies import has_idisplay
from .extend_docstring import extend_docstring

__all__ = ["extend_docstring", "get_bool_attribute", "get_attribute", "has_idisplay"]



""" set of target rates for binary classification """

from abc import ABC

from numpy import mean, median
from pandas import DataFrame, Series

from ..utils import TargetRate


class ContinuousTargetRate(TargetRate, ABC):
    """Continuous target rate class."""

    __name__ = "continuous_target_rate"

    def compute(self, xagg: Series) -> DataFrame:
        """Computes the target rate.

        Parameters
        ----------
        xagg : DataFrame
            A crosstab.

        Returns
        -------
        Series
            Target rate.
        """
        # checking for an xtab
        if xagg is not None:
            # frequency per modality
            frequency = xagg.apply(len) / xagg.apply(len).sum()

            # computing target rate
            return DataFrame({self.__name__: self._compute(xagg), "frequency": frequency})
        return None


class TargetMean(ContinuousTargetRate):
    """Mean target rate class."""

    __name__ = "target_mean"

    def _compute(self, xagg: DataFrame) -> Series:
        """Computes the mean target rate."""
        return xagg.apply(mean)


class TargetMedian(ContinuousTargetRate):
    """Median of target per class."""

    __name__ = "target_median"

    def _compute(self, xagg: DataFrame) -> Series:
        """Computes the mean target rate."""
        return xagg.apply(median)


# class TargetVariance(ContinuousTargetRate):
#     """Variance of target per class."""

#     __name__ = "target_variance"

#     def _compute(self, xagg: DataFrame) -> Series:
#         """Computes the mean target rate."""
#         return xagg.apply(var)


# class TargetStd(ContinuousTargetRate):
#     """Std of target per class."""

#     __name__ = "target_std"

#     def _compute(self, xagg: DataFrame) -> Series:
#         """Computes the mean target rate."""
#         return xagg.apply(std)


# class TargetIqr(ContinuousTargetRate):
#     """IQR of target per class."""

#     __name__ = "target_iqr"

#     def _compute(self, xagg: DataFrame) -> Series:
#         """Computes the mean target rate."""
#         return xagg.apply(iqr)


# class TargetRange(ContinuousTargetRate):
#     """Range of target per class."""

#     __name__ = "target_range"

#     def _compute(self, xagg: DataFrame) -> Series:
#         """Computes the mean target rate."""
#         return xagg.apply(lambda x: x.max() - x.min())



""" Defines a categorical feature"""

from abc import abstractmethod

from numpy import floating, integer
from pandas import DataFrame, Series, notna, unique

from ..utils.base_feature import BaseFeature
from ..utils.grouped_list import GroupedList


class QualitativeFeature(BaseFeature):
    """Defines a qualitative feature"""

    __name__ = "Qualitative"
    is_qualitative = True

    # def _update_value_per_label(self, raw_labels: list[str]) -> None:
    #     """updates value per label and label per value"""
    #     self.value_per_label = {}
    #     for value, label, raw_label in zip(self.values, self._labels, raw_labels):
    #         # updating label_per_value
    #         for grouped_value in self.values.get(value):
    #             self.label_per_value.update({grouped_value: label})

    #         # updating value_per_label
    #         self.value_per_label.update({label: raw_label})

    def fit(self, X: DataFrame, y: Series = None) -> None:
        """TODO fit stats"""

        # checking for feature's unique non-nan values
        sorted_unique_values = nan_unique(X[self.version], sort=True)

        # checking that feature is not ordinal (already set values)
        if self.values is None:
            # initiating feature with its unique non-nan values
            self.update(GroupedList(sorted_unique_values))

        # checking that raw order has not been set (also useful when loading from json)
        if len(self.raw_order) == 0:
            # saving up number ordering for labeling
            self.raw_order = [self.values.get_group(value) for value in sorted_unique_values]

        # fitting BaseFeature
        super().fit(X, y)

        # class-specific checking for unexpected values
        self.check_values(X)

    def check_values(self, X: DataFrame) -> None:
        """checks for unexpected values from unique values in DataFrame"""

        # computing unique labels in dataframe
        unique_labels = unique(X[self.version])

        # converting to labels
        unique_values = unique_labels[:]
        if not self.ordinal_encoding:
            unique_values = [self.value_per_label.get(label, label) for label in unique_labels]

        # unexpected values for this feature
        unexpected = [
            value
            for value in unique_values
            if not self.values.contains(value) and notna(value) and value != self.nan
        ]
        if len(unexpected) > 0:
            # feature does not have a default value
            if not self.has_default:
                raise ValueError(f"[{self}] Unexpected values: {str(list(unexpected))}")

            # feature has default value:
            # adding unexpected value to list of known values
            for unexpected_value in unexpected:
                self.values.append(unexpected_value)

            # adding unexpected to default
            default_group = self.values.get_group(self.default)
            self.group(unexpected, default_group)

        super().check_values(X)

    def make_labels(self) -> GroupedList:
        """gives labels per values"""

        # iterating over each value and there content
        labels = []
        for group, content in self.content.items():
            # formatting label
            labels += [self._format_modalities(group, content)]

        return GroupedList(labels)

    def _make_summary(self):
        """returns summary of feature's values' content"""
        # iterating over each value
        summary = []
        for group, values in self.content.items():
            # getting group label
            group_label = self.label_per_value.get(group)

            # Qualtiative features: filtering out numbers
            values = [value for value in values if isinstance(value, str)]

            # if there is only one value converting to str
            if len(values) == 1:
                values = values[0]

            # adding group summary
            summary += [{"feature": str(self), "label": group_label, "content": values}]

        # adding statistics and history
        return self._add_statistics_to_summary(summary)

    def _specific_update(self, values: GroupedList, convert_labels: bool = False) -> None:
        """update content of values specifically per feature type"""

        # no values have been set
        if not convert_labels and self.values is None:
            self.values = values

        # values are not labels
        elif not convert_labels:
            # updating: iterating over each grouped values
            for kept_value, grouped_values in values.content.items():
                self.values.group(grouped_values, kept_value)

        # values are labels -> converting them back to values
        else:
            # iterating over each grouped values
            for kept_label, grouped_labels in values.content.items():
                # converting labels to values
                kept_value = self.value_per_label.get(kept_label)
                grouped_values = [self.value_per_label.get(label) for label in grouped_labels]

                # checking that kept values exists
                if kept_label not in self.value_per_label:
                    raise AttributeError(
                        f"{self} no {kept_label}, in value_per_label: {self.value_per_label}"
                    )

                # checking that grouped values exists
                for grouped_value, grouped_label in zip(grouped_values, grouped_labels):
                    if grouped_value is None:
                        print(
                            f"{self} no {grouped_label}, in value_per_label: {self.value_per_label}"
                        )

                # choosing which value to keep: getting group of kept_value
                kept_value = self.values.get_group(kept_value)
                # TODO force kept_value to != self.nan like in quantitative feature?

                # keeping only values not already grouped with kept_value
                grouped_values = [
                    self.values.get_group(value)
                    for value in grouped_values
                    if self.values.get_group(value) != kept_value
                ]

                # deduplicating
                grouped_values = [
                    value
                    for num, value in enumerate(grouped_values)
                    if value not in grouped_values[num + 1 :]
                ]

                # updating values if any to group
                if len(grouped_values) > 0:
                    # if ordinal_encoding, converting values to unique values
                    if self.ordinal_encoding:
                        r_value_per_label = {
                            v: self.values[k] for k, v in self.value_per_label.items()
                        }
                        grouped_values = [r_value_per_label[value] for value in grouped_values]

                    # grouping values
                    self.values.group(grouped_values, kept_value)

                # updating statistics
                self._update_statistics_value(kept_label, kept_value)

    @abstractmethod
    def _specific_formatting(self, ordered_content: list[str]) -> str:
        """specific label formatting"""

    def _format_modalities(self, group: str, content: list[str]) -> list[str]:
        """Formats a list of float quantiles into a list of boundaries.

        Rounds quantiles to the closest power of 1000.

        Parameters
        ----------
        a_list : list[float]
            Sorted list of quantiles to convert into string

        Returns
        -------
        list[str]
            List of boundaries per quantile
        """

        # ordering content as per original ordering (removes DEFAULT and NAN)
        ordered_content = [
            value
            for value in self.raw_order
            if value in content
            # removing nan
            and value != self.nan
            # removing floats
            and not isinstance(value, floating) and not isinstance(value, float)
            # removing ints
            and not isinstance(value, integer) and not isinstance(value, int)
        ]
        # removing duplicates
        ordered_content = list(dict.fromkeys(ordered_content))

        # building label from ordered content
        if len(ordered_content) == 0:
            label = group
        elif len(ordered_content) == 1:
            label = ordered_content[0]
        else:
            label = self._specific_formatting(ordered_content)

        # adding nans
        if self.nan in content and label != self.nan:  # and self.nan not in label:
            label += f", {self.nan}"

        return label


def nan_unique(x: Series, sort: bool = False) -> list[str]:
    """Unique non-NaN values.

    Parameters
    ----------
    x : Series
        Values to be deduplicated.
    sorted : boolean, optionnal
        Whether or not to sort unique by appearance.

    Returns
    -------
    list[str]
        List of unique non-nan values
    """

    # unique values not sorted
    if sort:
        uniques = unique(x)

    # sorting unique values
    else:
        uniques = list(x.value_counts(sort=True, ascending=False).index)

    # filtering out nans
    uniques = [value for value in uniques if notna(value)]

    return uniques


def get_qualitative_features(features: list[BaseFeature]) -> list[QualitativeFeature]:
    """returns qualitative features amongst provided features"""
    return [feature for feature in features if feature.is_qualitative]



""" tools for docstrings"""

from typing import Callable


class extend_docstring:
    """Used to extend a Child's method docstring with the Parent's method docstring"""

    def __init__(self, method: Callable, append: bool = True, exclude: list[str] = None) -> None:
        self.doc = method.__doc__
        self.append = append
        self.exclude = exclude or []

    @staticmethod
    def split_sections(docstring: str) -> dict[str, list[str]]:
        """Split docstring into sections by double newlines"""
        # splitting parameters
        parameters = docstring.split("\n\n")

        # removing section names
        sections = {}
        current_section = ""
        section_params = []
        current_param = ""
        for param in parameters:
            # checking for section name
            if "------" in param:
                # storing current section
                if current_param:
                    section_params.append(current_param)
                    sections[current_section] = section_params
                    current_param = ""
                # resetting current section
                current_section = param
                section_params = []

            # it is a parameter
            else:
                # checking for parameter name
                if " : " in param:
                    # storing current param if any
                    if current_param:
                        section_params.append(current_param)

                    # resetting current param
                    current_param = param

                # adding to current param
                else:
                    if current_param:
                        current_param = "\n\n".join([current_param, param])
                    else:
                        current_param = param
        # storing last section
        if current_param:
            section_params.append(current_param)
            sections[current_section] = section_params

        return sections

    def filter_sections(self, sections: dict[str, list[str]]) -> str:
        """Filter sections based on parameters to exclude"""

        # removing parameters to exclude and converting back to docstring
        docstring = ""
        for section, section_params in sections.items():
            # adding section name
            docstring += section + "\n\n"

            # iterating over section parameters
            for param in section_params:
                # checking if parameter should be excluded
                if not any(exclude + " :" in param for exclude in self.exclude):
                    # adding parameters
                    docstring += param + "\n\n"

        return docstring

    def process_docstring(self) -> str:
        """Process the docstring to exclude specified parameters"""
        if not self.doc:
            return ""
        sections = self.split_sections(self.doc)
        filtered_sections = self.filter_sections(sections)
        return filtered_sections

    def __call__(self, function: Callable) -> Callable:
        """Extend the docstring of the function with the one from the method"""
        if self.doc is not None:
            self.doc = self.process_docstring()
            doc = function.__doc__
            function.__doc__ = self.doc
            if doc is not None:
                if self.append:
                    function.__doc__ = function.__doc__ + doc
                else:
                    function.__doc__ = doc + function.__doc__
        return function



"""Base tools to build simple buckets out of Quantitative and Qualitative features
for a binary classification model.
"""

from typing import Any, Union

from numpy import ndarray, sort
from pandas import isna


class GroupedList(list):
    """An ordered list that's extended with a per-value content dict."""

    def __repr__(self) -> str:
        return f"GroupedList({super().__repr__()})"

    def __init__(self, iterable: Union[ndarray, dict, list, tuple] = ()) -> None:
        """
        Parameters
        ----------
        iterable : Union[ndarray, dict, list, tuple], optional
            List-like or :class:`GroupedList`, by default ``()``
        """
        # case -1: iterable is an array
        if isinstance(iterable, ndarray):
            iterable = list(iterable)

        # case 0: iterable is the content dict
        if isinstance(iterable, dict):
            # storing ordered keys of the dict
            keys = list(iterable)

            # storing the values content per key
            self.content = dict(iterable.items())

            # checking that all values are only present once
            all_values = [value for values in iterable.values() for value in values]
            if not len(list(set(all_values))) == len(all_values):
                raise ValueError(f"[{self}] A value is present in several keys (groups)")

            # adding key to itself if it's not present in an other key
            keys_copy = keys[:]  # copying initial keys
            for key in keys_copy:
                # checking that the value is not comprised in an other key
                all_values = [
                    val
                    for iter_key, values in iterable.items()
                    for val in values
                    if key != iter_key
                ]
                if key not in all_values:
                    # checking that key is missing from its values
                    if key not in iterable[key]:
                        self.content.update({key: self.content[key] + [key]})
                # the key already is in another key (and its values are empty)
                # the key as already been grouped
                else:
                    self.content.pop(key)
                    keys.remove(key)

            # initiating the list with those keys
            super().__init__(keys)

        # case 1: copying a GroupedList
        elif hasattr(iterable, "content"):
            # initiating the list with the provided list of keys
            super().__init__(iterable)

            # copying values associated to keys
            self.content = dict(iterable.content.items())

        # case 2: initiating GroupedList from a list
        elif isinstance(iterable, list):
            # initiating the list with the provided list of keys
            super().__init__(iterable)

            # initiating the values with the provided list of keys
            self.content = {v: [v] for v in iterable}

    # def __getitem__(self, key):
    #     """Returns the value content in the key

    #     Parameters
    #     ----------
    #     key : Any
    #         Key to search for

    #     Returns
    #     -------
    #     list[Any]
    #         Values content in key

    #     Raises
    #     ------
    #     KeyError
    #         If key is not found in Group
    #     """

    #     if key in self.content:
    #         return self.content[key]

    #     raise KeyError(f"Key {key} not found in GroupedList")

    @property
    def values(self) -> list[str]:
        """All values content in all groups

        Returns
        -------
        list[str]
            List of all values in the GroupedList
        """
        return [value for values in self.content.values() for value in values]

    def get(self, key: Any, default: Any = None) -> list[Any]:
        """List of values content in key

        Parameters
        ----------
        key : Any
            Group.
        default : Any, optional
            Value to return if key was not found, by default None

        Returns
        -------
        list[Any]
            Values content in key
        """
        # default to fing an element
        default_value = []
        if default is not None:
            if isinstance(default, list):
                default_value = default
            else:
                default_value = [default]
        found = self.content.get(key, default_value)

        return found

    def sanity_check(self) -> None:
        """raises ValueError if there is an issue with the any element of the GroupedList"""
        # each element of the list should be in its elements
        for key, values in self.content.items():
            if key not in values:
                raise ValueError(f"[{self}] Missing group leader {key} in its values: {values}")

        # an element can not be in several groups
        all_values = self.values
        if len(set(all_values)) != len(all_values):
            raise ValueError(f"[{self}] Some values are in several groups")

        # checking that element of the list are keys of the content dict
        if not len(self) == len(self.content):
            raise ValueError(
                f"[{self}] Keys missing from content dict or element missing from list"
            )
        if any(
            list_element != dict_key
            for list_element, dict_key in zip(self, list(self.content.keys()))
        ):
            raise ValueError(f"[{self}] Not the same ordering between list and content dict")

    def _group_single_value(self, discarded: Any, kept: Any) -> None:
        """Groups the discarded value with the kept value

        Parameters
        ----------
        discarded : Any
            Value to be grouped into the key ``kept``.
        kept : Any
            Key value in which to group ``discarded``.
        """

        # checking that those values are distinct
        if not is_equal(discarded, kept):
            # checking that those values exist in the list
            if discarded not in self:
                raise ValueError(f"[{self}] {discarded} not in list")
            if kept not in self:
                raise ValueError(f"[{self}] {kept} not in list")

            # accessing values content in each value
            content_discarded = self.content.get(discarded)
            content_kept = self.content.get(kept)

            # updating content dict
            self.content.update({kept: content_discarded + content_kept, discarded: []})

            # removing discarded from the list
            self.remove(discarded)

    def group(self, to_discard: Union[list[str], str], to_keep: str) -> None:
        """Groups the discarded value with the kept value

        Parameters
        ----------
        to_discard : list[Any] | Any
            Values to be grouped into the key ``to_keep``.
        to_keep : Any
            Key value in which to group ``to_discard`` values.
        """
        # case of single value to discard
        if not isinstance(to_discard, list):
            self._group_single_value(to_discard, to_keep)

        # list of values to group
        else:
            for discarded in to_discard:
                self._group_single_value(discarded, to_keep)

        # sanity check after modification
        self.sanity_check()

    def append(self, new_value: Any) -> None:
        """Appends a new_value to the GroupedList

        Parameters
        ----------
        new_value : Any
            New key to be added.
        """

        # checking for already existing values
        if new_value in self.values:
            raise ValueError(f"- [{self}] Value {new_value} already in list!")

        # adding value to list and dict
        self += [new_value]
        self.content.update({new_value: [new_value]})

        # sanity check after modification
        self.sanity_check()

    def update(self, new_value: dict[Any, list[Any]]) -> None:
        """Updates the GroupedList via a dict

        Parameters
        ----------
        new_value : dict[Any, list[Any]]
            Dict of key, values to updated ``content`` dict
        """

        # should provide a dict of lists
        if not isinstance(new_value, dict) or not all(
            isinstance(value, list) for value in new_value.values()
        ):
            raise ValueError(f"[{self}] Provide a dictionnary of lists (values)")

        # adding missing keys to there list of values
        for key, value in new_value.items():
            if key not in value:
                new_value.update({key: value + [key]})

        # adding keys to the order if they are new values
        self += [key for key, _ in new_value.items() if key not in self]

        # updating content according to new_value
        self.content.update(new_value)

        # sanity check after modification
        self.sanity_check()

    def sort(self) -> "GroupedList":
        """Sorts the values of the list and dict (if any, NaNs are last).

        Returns
        -------
        GroupedList
            Sorted GroupedList
        """
        # str values
        keys_str = [key for key in self if isinstance(key, str)]

        # non-str values
        keys_float = [key for key in self if not isinstance(key, str)]

        # sorting and merging keys
        keys = list(sort(keys_str)) + list(sort(keys_float))

        # recreating an ordered GroupedList
        sorted_list = GroupedList({k: self.get(k) for k in keys})

        return sorted_list

    def sort_by(self, ordering: list[Any]) -> "GroupedList":
        """Sorts the values of the list and dict according to ``ordering``, if any,
        NaNs are the last.

        Parameters
        ----------
        ordering : list[Any]
            Order used for ordering of the list of keys.

        Returns
        -------
        GroupedList
            Sorted GroupedList
        """

        # checking that all values are given an order
        if not all(o in self for o in ordering):
            raise ValueError(
                f"[{self}] Unknown values in ordering: "
                f"{str([v for v in ordering if v not in self])}"
            )
        if not all(s in ordering for s in self):
            raise ValueError(
                f"[{self}] Missing value from ordering:"
                f" {str([v for v in self if v not in ordering])}"
            )

        # ordering the content
        sorted_list = GroupedList({k: self.get(k) for k in ordering})

        return sorted_list

    def remove(self, value: Any) -> None:
        """Removes a value from the GroupedList

        Parameters
        ----------
        value : Any
            value to be removed
        """
        super().remove(value)
        self.content.pop(value)

        # sanity check after modification
        self.sanity_check()

    def pop(self, idx: int) -> None:
        """Pop a value from the GroupedList by index

        Parameters
        ----------
        idx : int
            Index of the value to be popped out
        """
        value = self[idx]
        self.remove(value)

    def get_group(self, value: Any) -> Any:
        """Returns the key (group) containing the specified value

        Parameters
        ----------
        value : Any
            Value for which to find the group.

        Returns
        -------
        Any
            Corresponding key (group)
        """

        found = [
            key
            for key, values in self.content.items()
            if any(is_equal(value, elt) for elt in values)
        ]

        if any(found):
            return found[0]

        # TODO expected behavior?
        # if not found, should return itself by default
        return value

    def contains(self, value: Any) -> bool:
        """Checks if a value is content in any group, also matches NaNs.

        Parameters
        ----------
        value : Any
            Value to search for

        Returns
        -------
        bool
            Whether the value is in the GroupedList
        """
        return any(is_equal(value, known) for known in self.values)

    def replace_group_leader(self, group_leader: Any, group_member: Any) -> None:
        """Replaces a group_leader by one of its group_members

        Parameters
        ----------
        group_leader : Any
            One of the list's values (``GroupedList.content.keys()``)
        group_member : Any
            One of the dict's values for specified group_leader
            (``GroupedList.content[group_leader]``)
        """
        # checking that group_member is in group_leader
        if group_member not in self.content[group_leader]:
            raise ValueError(f"[{self}] {group_member} is not in {group_leader}")

        # replacing in the list
        group_idx = self.index(group_leader)
        self[group_idx] = group_member

        # replacing in the dict
        self.content.update({group_member: self.content[group_leader][:]})
        self.content.pop(group_leader)

        # TODO check if this has a purpose
        # sorting things up
        # self.sort_by(self)


def is_equal(a: Any, b: Any) -> bool:
    """Checks if a and b are equal (NaN insensitive)"""

    # default equality
    equal = a == b

    # Case where a and b are NaNs
    if isna(a) and isna(b):
        equal = True

    return equal



"""Tools to select the best Quantitative and Qualitative features."""

from abc import ABC, abstractmethod
from math import ceil
from random import seed, shuffle
from typing import Union

from pandas import DataFrame, Series

from ...features import BaseFeature, Features
from ...utils import get_bool_attribute, has_idisplay
from ..filters import BaseFilter
from ..measures import BaseMeasure
from .pretty_print import format_ranked_features, prettier_measures

# trying to import extra dependencies
_has_idisplay = has_idisplay()
if _has_idisplay:
    from IPython.display import display_html


class BaseSelector(ABC):
    """A pipeline of measures to perform a feature pre-selection that maximizes association
    with a binary target.

    Examples
    --------
    See `Selectors examples <https://autocarver.readthedocs.io/en/latest/index.html>`_
    """

    __name__ = "BaseSelector"

    def __init__(
        self,
        features: Features,
        n_best_per_type: int,
        # *,
        # measures: list[BaseMeasure] = None,
        # filters: list[BaseFilter] = None,
        # max_num_features_per_chunk: int = 100,
        **kwargs,
    ) -> None:
        # quantitative_measures : list[Callable], optional
        #     List of association measures to be used for ``quantitative_features``.
        #     :ref:`Implemented measures <Measures>` are:

        #     * For association evaluation:
        #         * :ref:`Kruskal-Wallis' H <kruskal>` (default)
        #         * :ref:`R`
        #     * For outlier detection:
        #         * :ref:`Standard score <zscore>`
        #         * :ref:`Interquartile range <iqr>`

        # qualitative_measures : list[Callable], optional
        #     List of association measures to be used for ``qualitative_features``.
        #     :ref:`Implemented measures <Measures>` are:

        #     * For association evaluation:
        #         * :ref:`Pearson's chi² <chi2>`
        #         * :ref:`cramerv`
        #         * :ref:`tschuprowt` (default)

        # quantitative_filters : list[Callable], optional
        #     List of filters to be used for ``quantitative_features``.
        #     :ref:`Implemented filters <Filters>` are:

        #     * For linear correlation:
        #         * :ref:`pearson_filter`
        #         * :ref:`Spearman's rho <spearman_filter>` (default)

        # qualitative_filters : list[Callable], optional
        #     List of filters to be used for ``qualitative_features``.
        #     :ref:`Implemented filters <Filters>` are:

        #     * For correlation:
        #         * :ref:`cramerv_filter`
        #         * :ref:`tschuprowt_filter` (default)

        # **kwargs
        #     Allows one to set thresholds for provided ``quantitative_measures``/
        #     ``qualitative_measures`` and ``quantitative_filters``/``qualitative_filters``
        #     (see :ref:`Measures` and :ref:`Filters`) passed as keyword arguments.
        """
        Parameters
        ----------
        features : Features
            A set of :class:`Features` to select from

        n_best_per_type : int
            Number of quantitative and/or qualitative :class:`Features` to  select


        Keyword Arguments
        -----------------
        measures : list[BaseMeasure], optional
            List of association measures to be used, by default ``None``.

            Selects :attr:`n_best_per_type` features for each measure provided.
            Implemented measures are:

            * :class:`QuantitativeFeature`: see available :ref:`QuantiMeasures`
            * :class:`QualitativeFeature`: see available :ref:`QualiMeasures`

        filters : list[BaseFilter], optional
            List of filters to be used, by default ``None``.

            Filters out features that do not pass the threshold of each filter.
            Implemented filters are:

            * :class:`QuantitativeFeature`: see available :ref:`QuantiFilters`
            * :class:`QualitativeFeature`: see available :ref:`QualiFilters`

        max_num_features_per_chunk : int, optional
            Maximum number of features per chunk, by default ``100``.

            Chunking is used to speed up the selection process for large numbers
            of :class:`Features`.

            1. :class:`Features` are split in ``n_chunks`` of :attr:`max_num_features_per_chunk`
            2. ``n_best_per_type//n_chunks`` of each chunk are selected
            3. best features are selected from the remaining features

        verbose : bool, optional
            * ``True``, without ``IPython``: prints raw statitics
            * ``True``, with ``IPython``: prints HTML statistics, by default ``False``
        """
        # TODO print a config that says what the user has selected (measures+filters)
        # features and values
        self.features = features
        if isinstance(features, list):
            self.features = Features(features)

        # number of features selected
        self.n_best_per_type = n_best_per_type
        if not (0 < int(self.n_best_per_type) <= len(self.features)):
            raise ValueError("Must set 0 < n_best_per_type <= len(features)")

        # feature sample size per iteration
        # maximum number of features per chunk
        self.max_num_features_per_chunk = kwargs.get("max_num_features_per_chunk", 100)
        if self.max_num_features_per_chunk <= 2:
            raise ValueError("Must set 2 < max_num_features_per_chunk")

        # random state for chunk shuffling
        self.random_state = kwargs.get("random_state", 0)

        # adding default measures
        self.measures = self._initiate_measures(kwargs.get("measures"))

        # adding default filter
        self.filters = self._initiate_filters(kwargs.get("filters"))

        # whether to print info
        self.verbose = get_bool_attribute(kwargs, "verbose", False)
        self._message = ""

        # target name
        self.target_name = None

    def __repr__(self) -> str:
        """Returns the name of the selector"""
        return self.__name__

    @property
    def pretty_print(self) -> bool:
        """Returns the pretty_print attribute"""
        return self.verbose and _has_idisplay

    @abstractmethod
    def _initiate_measures(self, requested_measures: list[BaseMeasure] = None) -> list[BaseMeasure]:
        """initiates the list of measures with default values"""
        return requested_measures

    @abstractmethod
    def _initiate_filters(self, requested_filters: list[BaseFilter] = None) -> list[BaseFilter]:
        """initiates the list of measures with default values"""
        return requested_filters

    def _select_quantitatives(
        self, quantitatives: list[BaseFeature], X: DataFrame, y: Series
    ) -> list[BaseFeature]:
        """Selects the best quantitative features"""

        # checking for quantitative features before selection
        best_quantitatives: list[BaseFeature] = []
        if len(quantitatives) > 0:
            # getting qualitative measures and filters
            measures = get_quantitative_metrics(self.measures)
            filters = get_quantitative_metrics(self.filters)

            # setting message for pretty print
            self._message = "Quantitative "

            # getting best features
            best_quantitatives = self._select_features(quantitatives, X, y, measures, filters)

        return best_quantitatives

    def _select_qualitatives(
        self, qualitatives: list[BaseFeature], X: DataFrame, y: Series
    ) -> list[BaseFeature]:
        """Selects the best qualitative features"""

        # checking for qualitative features before selection
        best_qualitatives: list[BaseFeature] = []
        if len(qualitatives) > 0:
            # getting qualitative measures and filters
            measures = get_qualitative_metrics(self.measures)
            filters = get_qualitative_metrics(self.filters)

            # setting message for pretty print
            self._message = "Qualitative "

            # getting best features
            best_qualitatives = self._select_features(qualitatives, X, y, measures, filters)

        return best_qualitatives

    def _initiate_features_measures(
        self, features: list[BaseFeature], remove_default: bool = True
    ) -> None:
        """initiates the list of measures with default values"""
        # iterating over each feature
        for feature in features:
            # removing all measures
            if remove_default:
                feature.measures = {}
                feature.filters = {}

            # keeping only default measures
            else:
                remove_non_default_metrics_from_features(feature)

    def select(self, X: DataFrame, y: Series) -> Features:
        """Selects the :attr:`n_best_per_type` :class:`Features` of ``X``

        Parameters
        ----------
        X : DataFrame
            Dataset to determine optimal features.

        y : Series
            Target with wich the association is evaluated.

        Returns
        -------
        Features
            Selected :class:`Features`
        """

        # getting target name
        if isinstance(y, Series):
            self.target_name = y.name

        # initiating features measures and filters
        self._initiate_features_measures(self.features, remove_default=True)

        # checking for quantitative and qualitative features
        features = get_typed_features(self.features)

        # selecting quantitative features
        best_features = self._select_quantitatives(features.get("quantitatives"), X, y)

        # selecting qualitative features
        best_features += self._select_qualitatives(features.get("qualitatives"), X, y)

        # converting to Features
        if len(best_features) > 0:
            return Features(best_features)
        return best_features

    def _select_features(
        self,
        features: list[BaseFeature],
        X: DataFrame,
        y: Series,
        measures: list[BaseMeasure],
        filters: list[BaseFilter],
    ) -> list[BaseFeature]:
        """selects amongst features"""

        # apply default measures to features
        apply_measures(features, X, y, measures, default_measures=True)

        # apply default filters to features
        features = apply_filters(features, X, filters, default_filters=True)

        # getting non-default measures and filters
        measures = remove_default_metrics(measures)
        filters = remove_default_metrics(filters)

        # splitting features in chunks and getting best per-chunk set of features
        return self._get_best_features_across_chunks(features, X, y, measures, filters)

    def _get_best_features_across_chunks(
        self,
        features: list[BaseFeature],
        X: DataFrame,
        y: Series,
        measures: list[BaseMeasure],
        filters: list[BaseFilter],
    ) -> list[BaseFeature]:
        """gets best features per chunk of maximum size as set by max_num_features_per_chunk"""

        # too many features: chunking and selecting best amongst chunks
        if len(features) > self.max_num_features_per_chunk:
            # making random chunks of features
            chunks = make_random_chunks(
                features, self.max_num_features_per_chunk, self.random_state
            )

            # iterating over each feature samples
            features: list[BaseFeature] = []
            for n_chunk, chunk in enumerate(chunks):
                # fitting association on features
                features += get_best_features(
                    chunk, X, y, measures, filters, max(int(self.n_best_per_type // len(chunks)), 1)
                )

                # printing association
                self._print_measures(chunk, f"from chunk {n_chunk}/{len(chunks)}")

        # initiating features measures and filters for final prediction
        self._initiate_features_measures(features, remove_default=False)

        # selecting from remaining features
        best_features = get_best_features(features, X, y, measures, filters, self.n_best_per_type)

        # printing association
        self._print_measures(features)

        return best_features

    def _print_measures(self, features: list[BaseFeature], message: str = "") -> None:
        """Prints crosstabs' target rates and frequencies per modality, in raw or html format

        Parameters
        ----------
        xagg : Union[DataFrame, Series]
            Train crosstab
        xagg_dev : Union[DataFrame, Series]
            Dev crosstab, by default None
        pretty_print : bool, optional
            Whether to output html or not, by default False
        """
        if self.verbose:
            # formatting measures to print
            formatted_measures = format_ranked_features(features)

            if not formatted_measures.empty:
                print(f" [{self.__name__}] Selected {self._message}Features {message}")

            if not self.pretty_print:  # no pretty hmtl printing
                self._print_raw(formatted_measures)
            else:  # pretty html printing
                self._print_html(formatted_measures)

    def _print_raw(self, formatted_measures: DataFrame) -> None:
        """Prints raw XAGG DataFrames."""
        print(formatted_measures, "\n")

    def _print_html(self, formatted_measures: DataFrame) -> None:
        """Prints XAGG DataFrames in HTML format."""
        # getting prettier xtabs
        nicer_association = prettier_measures(formatted_measures)

        # displaying html of colored DataFrame
        display_html(nicer_association, raw=True)  # pylint: disable=W0212


def get_typed_features(features: Features) -> dict[str, list[BaseFeature]]:
    """returns quantitative and qualitative features from list of features"""
    return {
        "quantitatives": [feature for feature in features if is_quantitative(feature)],
        "qualitatives": [feature for feature in features if is_qualitative(feature)],
    }


def is_quantitative(feature: BaseFeature) -> bool:
    """checks if feature is quantitative"""
    return feature.is_quantitative and not feature.is_fitted


def is_qualitative(feature: BaseFeature) -> bool:
    """checks if feature is qualitative"""
    return feature.is_qualitative or feature.is_fitted


def make_random_chunks(elements: list, max_chunk_sizes: int, random_state: int = None) -> list:
    """makes a specific number of random chunks of of a list"""

    # copying in order to not moidy initial list
    shuffled_elements = elements[:]

    # shuffling features to get random samples of features
    seed(random_state)
    shuffle(shuffled_elements)

    # number of chunks
    num_chunks = ceil(len(elements) / max_chunk_sizes)

    # getting size of each chunk
    chunk_size, remainder = divmod(len(elements), num_chunks)

    # getting all chunks
    chunks = []
    start = 0
    for i in range(num_chunks):
        end = start + chunk_size + (1 if i < remainder else 0)
        chunks.append(shuffled_elements[start:end])
        start = end

    return chunks


def get_qualitative_metrics(
    metrics: list[Union[BaseMeasure, BaseFilter]]
) -> list[Union[BaseMeasure, BaseFilter]]:
    """returns filtered list of measures/filters that apply on qualitative features"""
    return [metric for metric in metrics if metric.is_x_qualitative]


def get_quantitative_metrics(
    metrics: list[Union[BaseMeasure, BaseFilter]]
) -> list[Union[BaseMeasure, BaseFilter]]:
    """returns filtered list of measures/filters that apply on quantitative features"""
    return [metric for metric in metrics if metric.is_x_quantitative]


def get_default_metrics(
    metrics: list[Union[BaseMeasure, BaseFilter]]
) -> list[Union[BaseMeasure, BaseFilter]]:
    """returns filtered list of measures/filters that are default"""
    return [metric for metric in metrics if metric.is_default]


def remove_default_metrics(
    metrics: list[Union[BaseMeasure, BaseFilter]]
) -> list[Union[BaseMeasure, BaseFilter]]:
    """returns filtered list of measures/filters that are non-default"""
    return [metric for metric in metrics if not metric.is_default]


def remove_non_default_metrics_from_features(feature: BaseFeature) -> None:
    """removes default measures from feature"""
    # removing non default measures
    measures = dict(feature.measures)
    for measure_name, measure in feature.measures.items():
        if not measure.get("info").get("is_default"):
            measures.pop(measure_name)

    # removing non default filters
    filters = dict(feature.filters)
    for filter_name, measure in feature.filters.items():
        if not measure.get("info").get("is_default"):
            filters.pop(filter_name)

    # updating feature
    feature.measures = measures
    feature.filters = filters


def remove_duplicates(features: list[BaseFeature]) -> list[BaseFeature]:
    """removes duplicated features, keeping its first appearance"""
    return [features[i] for i in range(len(features)) if features[i] not in features[:i]]


def sort_features_per_measure(
    features: list[BaseFeature], measure: BaseMeasure
) -> list[BaseFeature]:
    """sorts features according to specified measure"""
    # checking if features are already ranked
    ranked = False
    for feature in features:
        if make_rank_name(measure) in feature.measures:
            ranked = True

    # setting reverse mode
    reverse = not measure.info.get("higher_is_better")
    if ranked:
        reverse = False

    # sorting features according to measure
    return sorted(features, key=lambda feature: get_feature_rank(feature, measure), reverse=reverse)


def get_feature_rank(feature: BaseFeature, measure: BaseMeasure) -> float:
    """gives rank of feature according to measure"""
    if make_rank_name(measure) not in feature.measures:
        return get_measure_value(feature, measure)
    return get_measure_rank(feature, measure)


def get_measure_rank(feature: BaseFeature, measure: BaseMeasure) -> int:
    """gives rank of feature according to measure"""
    return feature.measures.get(make_rank_name(measure)).get("value")


def get_measure_value(feature: BaseFeature, measure: BaseMeasure) -> float:
    """gives value of measure for specified feature"""
    return feature.measures.get(measure.__name__).get("value")


def apply_measures(
    features: list[BaseFeature],
    X: DataFrame,
    y: Series,
    measures: list[BaseMeasure],
    default_measures: bool = False,
) -> list[BaseMeasure]:
    """Measures association between columns of X and y, returns remaining_measures (not used)"""

    # keeping only default measures or non default measures
    used_measures = remove_default_metrics(measures)
    if default_measures:
        used_measures = get_default_metrics(measures)

    # iterating over each feature
    for feature in features:
        # iterating over each measure
        for measure in used_measures:
            # checking for mismatched data types
            check_measure_mismatch(feature, measure)

            # computing association for feature
            measure.compute_association(X[feature.version], y)

            # updating feature accordingly
            measure._update_feature(feature)


def apply_filters(
    features: list[BaseFeature],
    X: DataFrame,
    filters: list[BaseFilter],
    default_filters: bool = False,
) -> list[BaseFeature]:
    """Filters out too correlated features (least relevant first)"""

    # keeping only default filters or non default filters
    used_filters = remove_default_metrics(filters)
    if default_filters:
        used_filters = get_default_metrics(filters)

    # keeping track of remaining features
    filtered = features[:]

    # iterating over each filter
    for measure in used_filters:
        # checking for mismatched data types
        for feature in features:
            check_measure_mismatch(feature, measure)

        # applying filter
        filtered = measure.filter(X, filtered)

    return filtered


def check_measure_mismatch(feature: BaseFeature, measure: Union[BaseMeasure, BaseFilter]) -> None:
    """checks for mismatched data types between feature and measure"""
    if not (
        (is_quantitative(feature) and measure.is_x_quantitative)
        or (is_qualitative(feature) and measure.is_x_qualitative)
    ):
        raise TypeError(
            f"Type mismatch, provided feature {feature}, with {measure} that has "
            f"is_x_quantitative={measure.is_x_quantitative}"
        )


def get_best_features(
    features: list[BaseFeature],
    X: DataFrame,
    y: Series,
    measures: list[BaseMeasure],
    filters: list[BaseFilter],
    n_best: int,
) -> list[BaseFeature]:
    """gives best features according to provided measures"""

    # check for sortable measures
    if not all(measure.is_sortable for measure in measures):
        raise ValueError("All provided measures should be sortable")

    # applying measures to all features
    apply_measures(features, X, y, measures)

    # getting best feature for each sortable measure
    best_features = []
    for measure in measures:
        # getting best features according to measure
        best_features += select_with_measure(X, features, measure, filters, n_best)

    # deduplicating best features
    return remove_duplicates(best_features)


def select_with_measure(
    X: DataFrame,
    features: list[BaseFeature],
    measure: BaseMeasure,
    filters: list[BaseFilter],
    n_best: int,
) -> list[BaseFeature]:
    """Selects the ``n_best`` features of the DataFrame, by association with the target"""

    # sorting features according to measure
    sorted_features = sort_features_per_measure(features, measure)

    # reversing features
    sorted_features.reverse()

    # applying filters
    filtered_features = apply_filters(sorted_features, X, filters)

    # adding info to features
    for rank, feature in enumerate(filtered_features):
        feature.measures.update(make_rank_info(rank, measure, n_best, len(filtered_features)))

    # getting best features according to measure
    return select_from_rank(filtered_features, measure)


def select_from_rank(features: list[BaseFeature], measure: BaseMeasure) -> list[BaseFeature]:
    """Selects the ``n_best`` features of the DataFrame, by association with the target"""
    return [
        feature
        for feature in features
        if feature.measures.get(make_rank_name(measure)).get("valid")
    ]


def make_rank_name(measure: BaseMeasure) -> str:
    """makes a name for the rank info"""
    return f"{measure.__name__.replace('Measure', '')}Rank"


def make_rank_info(rank: int, measure: BaseMeasure, n_best: int, n_features: int) -> dict:
    """makes a dict with rank and measure info"""
    return {
        make_rank_name(measure): {
            "value": rank,
            "threshold": n_features - n_best,
            "valid": rank < n_best,
            "info": {"is_default": False, "higher_is_better": False},
        }
    }
